/******************************************************************************/
/*file rsa.c                      RSA functions                               */
/*                          coded by Daniele Sgandurra                        */
/******************************************************************************/


#include <stdio.h>
#include <time.h>
#include "cryptlib.h"
#include <openssl/bn.h>
#include <openssl/rsa.h>
#include <openssl/rand.h>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <errno.h>
#include "rsa.h"
#include "ssl.h"


/*per stampare a video i valori che compongono la chiave RSA "rsa"*/
void print_values(RSA *rsa)
{
     char *P, *Q, *N, *E, *D;/*i valori da stampare*/

     P = BN_bn2dec(rsa->p);
     Q = BN_bn2dec(rsa->q);
     N = BN_bn2dec(rsa->n);
     E = BN_bn2dec(rsa->e);
     D = BN_bn2dec(rsa->d);
     printf("p = %s\n", P);
     printf("q = %s\n", Q);
     printf("n = %s\n", N);
     printf("e = %s\n", E);
     printf("d = %s\n", D);
     
     OPENSSL_free(P);
     OPENSSL_free(Q);    
     OPENSSL_free(N);
     OPENSSL_free(E);
     OPENSSL_free(D);
}


/*genera una chiave RSA di "bits" bit a partire da "p_value" e "q_value" noti*/
RSA *RSA_generate_key_commutative(int bits, BIGNUM *p_value, BIGNUM *q_value)
{
     RSA *rsa=NULL; /*la chiave da generare*/
     BIGNUM *r0=NULL,*r1=NULL,*r2=NULL, *r3=NULL;/*utilizzati per generare la chiave*/
     int bitsp,bitsq,ok= -1;
     BN_CTX *ctx=NULL, *ctx2=NULL;
     char func_name[] = "RSA_generate_key_commutative";/*il nome della funzione*/

     ctx=BN_CTX_new();
     if (ctx == NULL) goto err;
     ctx2=BN_CTX_new();
     if (ctx2 == NULL) goto err;
     BN_CTX_start(ctx);
     r0 = BN_CTX_get(ctx);
     r1 = BN_CTX_get(ctx);
     r2 = BN_CTX_get(ctx);
     r3 = BN_CTX_get(ctx);
	 
     bitsp=(bits+1)/2;
     bitsq=bits-bitsp;
     rsa=RSA_new();
     if (rsa == NULL) goto err;

     /* set p */ 
     rsa->p=BN_new();
     if (rsa->p == NULL) goto err;
     if(BN_copy(rsa->p , p_value) == NULL) goto err;


     /* set q */ 
     rsa->q=BN_new();
     if (rsa->q == NULL) goto err;
     if(BN_copy(rsa->q , q_value) == NULL) goto err;


     /* calculate n */
     rsa->n=BN_new();
     if (rsa->n == NULL) goto err;
     if (!BN_mul(rsa->n,rsa->p,rsa->q,ctx)) goto err;

     /*calculate e */
     if (!BN_sub(r1,rsa->p,BN_value_one())) goto err;        /*r1 = p-1 */
     if (!BN_sub(r2,rsa->q,BN_value_one())) goto err;        /*r2 = q-1 */
     if (!BN_mul(r0,r1,r2,ctx)) goto err;    /* r0 = (p-1)(q-1) */
     
     /* generate e */
     for (;;)
     {
	  rsa->e = BN_new();
	  if(!BN_rand_range(rsa->e, rsa->n)) goto err; /*genero e minore di n*/
	  if (rsa->e == NULL) goto err;
	  if (!BN_gcd(r3, r0, rsa->e, ctx)) goto err; /*e non deve dividere r0 = (p-1)(q-1)*/
	  if (BN_is_one(r3)) break;
	  BN_free(rsa->e);
     }


     /* calculate d */
     rsa->d=BN_mod_inverse(NULL,rsa->e,r0,ctx2);     /* d */
     if (rsa->d == NULL) goto err;

 
     /* calculate d mod (p-1) */
     rsa->dmp1=BN_new();
     if (rsa->dmp1 == NULL) goto err;
     if (!BN_mod(rsa->dmp1,rsa->d,r1,ctx)) goto err;

     /* calculate d mod (q-1) */
     rsa->dmq1=BN_new();
     if (rsa->dmq1 == NULL) goto err;
     if (!BN_mod(rsa->dmq1,rsa->d,r2,ctx)) goto err;


     /* calculate inverse of q mod p */
     rsa->iqmp=BN_mod_inverse(NULL,rsa->q,rsa->p,ctx2);
     if (rsa->iqmp == NULL) goto err;

     ok=1;

 err:
     if(ok != 1)
     {
	  print_ssl_error(func_name, "RSA_generate_key_commutative");
     }
     if (ok == -1)
     {
	  RSAerr(RSA_F_RSA_GENERATE_KEY,ERR_LIB_BN);
	  ok=0;
     }
     BN_CTX_end(ctx);
     BN_CTX_free(ctx);
     BN_CTX_free(ctx2);
     if (!ok)
     {
	  if (rsa != NULL) RSA_free(rsa);
	  return(NULL);
     }
     else
	  return(rsa);
}


/*genera una chiave RSA con i valori di default*/
RSA *RSA_gen_key()
{	
     RSA *key_pair = RSA_generate_key(MODULUS_SIZE, PUBLIC_EXPONENT, NULL, NULL);
     return key_pair;
}


/*genera una chiave RSA con lo stesso modulo contenuto in "key"
  (per crittografia commutativa*/
RSA *RSA_gen_key_commutative(RSA *key)
{ 
     RSA *key_pair = RSA_generate_key_commutative(MODULUS_SIZE, key->p, key->q);
     return key_pair;
}


/*scrive le chiavi contenute in "rsa" sui file "pubname" per la chiave pubblica e 
  "privname" per quella privata, utilizzando per salvare la chiave privata la
  password "password"*/
int write_rsa_keys(char *pubname, char *privname, RSA *rsa, char *password)
{
     FILE *fd_pub;/*fd del file contenente la chiave pubblica*/
     FILE *fd_priv;/*fd del file contenente la chiave privata*/
     char func_name[] = "write_rsa_keys";/*il nome della funzione*/
  
     fd_pub = fopen(pubname,"w");
     if(fd_pub == NULL)
     {
	  printf("Cant open public key file %s\n", pubname);
	  perror("fopen");
	  return -1;
     }
     fd_priv = fopen(privname,"w");
     if(fd_priv == NULL)
     {
	  printf("Cant open private key file %s\n", privname);
	  perror("fopen");
	  return -1;
     }
     if(PEM_write_RSA_PUBKEY(fd_pub, rsa) < 0)
     {
	  print_ssl_error(func_name, "PEM_write_RSA_PUBKEY");
	  return -1;
     }

     if(PEM_write_RSAPrivateKey(fd_priv, rsa,EVP_des_ede3_cbc(), NULL, 
				0, NULL, password) < 0)
     {
	  print_ssl_error(func_name, "PEM_write_RSAPrivateKey");
	  return -1;
     }     
     fclose(fd_pub);
     fclose(fd_priv);
     
     return 0;
}


/*calcola la dimensione dei blocchi cifrati col padding, usando la chiave salvata
  in "pub_key_file", ed una stringa di lunghezza "length".
  Se la dimensione del modulo e' 1024 bits, per esempio, ritorna un multiplo di 128
  (dimensione di byte) */
int calculate_size(char *pub_key_file, int length)
{
     int size; /*il valore ritornato*/
     int key_size;/*la dimensione della chiave*/
     FILE *fd_pub;/*fd del file contenente la chiave pubblica*/
     RSA *pub_key=NULL;/*la chiave pubblica*/
     char func_name[] = "calculate_size";/*il nome della funzione*/

     if((fd_pub = fopen(pub_key_file,"r")) == NULL)
     {
	  printf("Error opening public key %s\n", pub_key_file);
	  exit(1);
     }
     if((pub_key = PEM_read_RSA_PUBKEY(fd_pub,NULL,0,NULL)) == NULL)
     {
	  print_ssl_error(func_name, "PEM_read_RSA_PUBKEY");
	  return -1;
     }
     key_size = RSA_size(pub_key); /*dimensione chiave*/
     size = (((length - 1) / key_size) + 1) * key_size; /*multiplo della chiave*/
     fclose(fd_pub);
     
     return size;
}


/*critta "from" di dimensione "length" utilizzando la chiave salvata su "pub_key_file"; 
  ritorna un array di unsigned char; funzione utilizzata la prima volta
  sul testo in chiaro*/
unsigned char *rsa_encrypt1(char *pub_key_file, char *from, int length)
{
     RSA *pub_key=NULL;/*la chiave RSA utilizzata*/
     unsigned char *text;/*buffer per contenere il testo da cifrare*/
     int inmax; /*dimensione chiave:specifica l'input massimo*/
     int text_length;/*lunghezza del testo crittato*/
     FILE *fd_pub;/*fd del file contenente la chiave pubblica*/
     int i;/*per scorrere l'array*/
     unsigned char *to; /*l'array che viene ritornato*/
     unsigned char *ptext, *pto; /*puntatori a text e to*/
     char func_name[] = "rsa_encrypt1";/*il nome della funzione*/     

     if((fd_pub = fopen(pub_key_file,"r")) == NULL)
     {
	  printf("[%s]: error opening public key %s\n", func_name, pub_key_file);
	  exit(1);
     }

     if((pub_key = PEM_read_RSA_PUBKEY(fd_pub,NULL,0,NULL)) == NULL)
     {
	  print_ssl_error(func_name, "PEM_read_RSA_PUBKEY");
	  return NULL;
     }
     inmax = RSA_size(pub_key); /*dimensione chiave*/
     text_length = (((length - 1) / inmax) + 1) * inmax; /*multiplo della chiave*/ 
     to =  malloc(text_length * sizeof(unsigned char)); 
     text = malloc(text_length * sizeof(unsigned char));
     for(i = 0; i < length; i++)
     {
	  text[i] = from[i];
     }
     if(length < text_length)
     {
	  text[length] = '\0';/* lo chiudo*/
     }
     for(i = length + 1; i < text_length; i++)
     {
	  text[i] = '0'; /*lo riempio di zero*/
     }
     length = 0;
     ptext = &text[0];
     pto = &to[0];
     while(length != text_length)
     {
	  ptext = &text[length];
	  pto = &to[length];
	  if((RSA_public_encrypt(inmax , ptext, pto, pub_key, RSA_NO_PADDING)) < 0)
	  {
	       print_ssl_error(func_name, "RSA_public_encrypt");
	       return NULL;
	  }
	  length += inmax;
     }
     free(text);
     fclose(fd_pub);

     return to;
}


/*critta "from" di dimensione "length" utilizzando la chiave salvata su "pub_key_file"; 
  ritorna un array di unsigned char: funzione utilizzata per crittare buffer gia' crittati*/
unsigned char *rsa_encrypt2(char *pub_key_file, unsigned char *from, int length)
{
     RSA *pub_key=NULL;/*la chiave RSA utilizzata*/
     int inmax; /*dimensione chiave:specifica l'input massimo*/
     FILE *fd_pub;/*fd del file contenente la chiave pubblica*/
     int i;/*per scorrere l'array*/
     unsigned char *to; /*l'array che viene ritornato*/
     unsigned char *pfrom, *pto; /*puntatori a text e to*/
     char func_name[] = "rsa_encrypt2";/*il nome della funzione*/     

     if((fd_pub = fopen(pub_key_file,"r")) == NULL)
     {
	  printf("[%s]: error opening public key [%s]\n", func_name, pub_key_file);
	  exit(1);
     }

     if((pub_key = PEM_read_RSA_PUBKEY(fd_pub,NULL,0,NULL)) == NULL)
     {
	  print_ssl_error(func_name, "PEM_read_RSA_PUBKEY");
	  return NULL;
     }
     inmax = RSA_size(pub_key); /*dimensione chiave*/
     to =  malloc(length * sizeof(unsigned char));
     i = 0;
     pfrom = &from[0];
     pto = &to[0];
     while(i != length)
     {
	  pfrom = &from[i];
	  pto = &to[i];
	  if((RSA_public_encrypt(inmax , pfrom, pto, pub_key, RSA_NO_PADDING)) < 0)
	  {
	       print_ssl_error(func_name, "RSA_public_encrypt");
	       return NULL;
	  }
	  i += inmax;
     }
     fclose(fd_pub);

     return to;
}


/*decritta "from" di lunghezza "size" nell'array "to", utilizzando la chiave privata
  salvata in "priv_key_file" con la password "password"*/
int rsa_decrypt(char *priv_key_file, unsigned char *from, unsigned char *to, 
		int size, char *password)
{
     RSA *priv_key=NULL;/*la chiave RSA utilizzata*/
     FILE *fd_priv;/*fd del file contenente la chiave privata*/
     int inmax; /*dimensione chiave:specifica l'input massimo*/
     int i = 0;/*per scorrere l'array*/
     unsigned char *pfrom, *pto;/*puntatori a from e to*/
     char func_name[] = "rsa_decrypt";/*il nome della funzione*/

     if((fd_priv = fopen(priv_key_file,"r")) == NULL)
     {
	  printf("[%s]: error opening private key %s\n", func_name, priv_key_file);
	  exit(1);
     }
     
     if((priv_key = PEM_read_RSAPrivateKey(fd_priv, NULL,0, password)) == NULL)
     {
	  print_ssl_error(func_name, "PEM_read_RSAPrivateKey");
	  return (-1);
     }
     pfrom = &from[0];
     pto = &to[0];
     inmax = RSA_size(priv_key); /*dimensione chiave*/
     if((size % inmax) != 0)
     {
	  printf("[%s]: \"size\" must be a multiple of RSA key size (%d bytes)\n", func_name, inmax);
	  return -1;
     }
     while(i != size)
     {
	  pfrom = &from[i];
	  pto = &to[i];
	  if((RSA_private_decrypt(inmax, pfrom, pto, priv_key, RSA_NO_PADDING)) < 0)
	  {
	       print_ssl_error(func_name, "RSA_private_decrypt");
	       return (-1);
	  }
	  i +=inmax;
     }
     fclose(fd_priv);

     return 0;
}
