/******************************************************************************/
/*file secure.c            SECURE MULTIPARTY COMPUTATION                      */
/*                                primitives                                  */
/*                          coded by Daniele Sgandurra                        */
/******************************************************************************/


/******************************************************************************/
/*                                                                            */
/*                                  INCLUDE                                   */
/*                                                                            */
/******************************************************************************/


#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <limits.h>
#include <sys/types.h>
#include <openssl/ssl.h>
#include <openssl/x509.h>
#include <openssl/rand.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <pthread.h>
#include "secure.h"
#include "ssl.h"
#include "rsa.h"


/******************************************************************************/
/*                                                                            */
/*                                   DEFINE                                   */
/*                                                                            */
/******************************************************************************/


#define MAX_FILENAME_LENGTH 100
#define NAME_LENGTH 50
#define BUF_LENGTH 300
#define BIGNUM_BUF_LENGTH 250
#define MSG_LENGTH 5
#define MAX_IP_LENGTH 16
#define MOD ULONG_MAX
#define SUM_PORT_NUMBER 8815
#define UNION_PORT_NUMBER 8816
#define INTER_PORT_NUMBER 8817
#define SCALAR_PORT_NUMBER 8818
#define SECURE_UNION 0
#define SECURE_INTERSECTION 1
#define SUM_CONNECT_TIMEOUT 20000
#define UNION_CONNECT_TIMEOUT 20000
#define SCALAR_CONNECT_TIMEOUT 3000
#define MIN_CLIENTS_NUM 2
#define SLEEP_TIME 10000
#define MAX_THREADS_NUMBER 1000
#define DEFAULT_BLOCK_LENGTH 1
#define ENCRYPT 1
#define DECRYPT 2
#define SEND_CRYPT 3
#define SEND_CLEAR 4
#define SEND_RSA 5
#define SEND_INTERSECTION 6
#define N_R_PROPORTION 3
#define DEFAULT_COMMUTATIVE_PASSWORD "secure_passwd"
#define CERTIFICATE_EXT ".certificate.pem"
#define PRIVATE_EXT ".private.pem"
#define CAFILE "certificate/CAfile.pem"
#define CERTIFICATE_DIR "certificate/"
#define COMM_PUB_KEY_EXT ".comm_public_key.pem"
#define COMM_PRI_KEY_EXT ".comm_private_key.pem"


/******************************************************************************/
/*                                                                            */
/*                              VARIABILI GLOBALI                             */
/*                                                                            */
/******************************************************************************/


/*variabili usate per sincronizzare i thread*/
pthread_mutex_t mutex_critt_params = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cond_critt_params = PTHREAD_COND_INITIALIZER;


/******************************************************************************/
/*                                                                            */
/*                                   TYPEDEF                                  */
/*                                                                            */
/******************************************************************************/


/*puntatore alla struttura sockaddr_in*/
typedef struct sockaddr_in *psockaddr_in;


/*serve alla connect timeout*/
typedef void (sighandler_t)(int);


/*struttura usata dai thread della secure_union_intersection per inviare i valori di RSA*/
typedef struct params_rsa_s
{
     SSL_CTX* ctx_client; /*il contesto  del client*/
     struct sockaddr *sock_address; /*il sockaddr del prossimo client*/
     BIGNUM *p; /*p per generare le chiavi commutative*/
     BIGNUM *q; /*q per generare le chiavi commutative*/
}param_rsa_s;


/*struttura usata dai thread della secure_union_intersection per ricevere i valori di RSA*/
typedef struct params_rsa_r
{
     SSL *ssl_server; /*il server ssl*/
     BIGNUM **p; /*p per generare le chiavi commutative*/
     BIGNUM **q;/*q per generare le chiavi commutative*/
}param_rsa_r;


/*struttura usata dai thread della secure_union_intersection per contenere i crittogrammi crittati da tutti*/
typedef struct params_union_intersection_receive_c
{
     SSL *ssl_server;/*il server ssl*/
     int tot;/*quanti client hanno gia' inviato i messaggi*/
     int max;/*quanti client devono spedire i messaggi*/
     unsigned long n_critt;/*numero totale di crittogrammi ricevuti*/
     unsigned char ***critt;/*i crittogrammi*/
     int len;/*lunghezza crittogrammi*/
}param_union_intersection_receive_c;


/*struttura usata dai thread della secure_union_intersection per ricevere e crittare i dati*/
typedef struct params_union_intersection_c
{
     int other_sockfd;/*il socket da cui ricevere*/
     SSL *ssl_server;/*il server ssl*/
     int len;/*lunghezza stringhe ricezione ed invio*/
     SSL_CTX *ctx_client;/*il contesto del client*/
     struct sockaddr *sock_address;/*il sockaddr del prossimo client*/
     struct sockaddr *server_sock_address;/*il sockaddr del client 0*/
     char *comm_public_key;/*chiave pubblica commutativa*/
     char **clear;/*usata solo la prima volta per contenere i messaggi in chiaro*/
     unsigned long n_clear;/*quanti messaggi in chiaro*/
     int max;/*quanti client devono crittare il messaggio*/
     struct params_union_intersection_receive_c *par_rec_c;/*serve unicamente al primo client*/
}param_union_intersection_c;


/*struttura usata dai thread della secure_union_intersection per ricevere e decrittare i dati*/
typedef struct params_union_d
{
     SSL *ssl_server;/*il server ssl*/
     int len;/*lunghezza stringhe ricezione ed invio*/
     SSL_CTX *ctx_client;/*il contesto del client*/
     struct sockaddr *sock_address;/*il sockaddr del prossimo client*/
     struct sockaddr **server_sock_address;/*i sockaddr di tutti i client (serve all'ultimo)*/
     char *comm_private_key;/*chiave privata commutativa*/
     int max;/*quanti client devono crittare il messaggio*/
     char *password;/*usata per leggere la chiave privata*/
     unsigned char ***clear;/*usata solo dall'ultimo*/
     unsigned long *n_clear;/*usata solo dall'ultimo*/
}param_union_d;


/*struttura usata dai thread della secure_union_intersection del primo host per inviare le stringhe da decrittare*/
typedef struct params_critt_s
{
     int len;/*lunghezza stringhe ricezione ed invio*/
     SSL_CTX *ctx_client;/*il contesto del client*/
     struct sockaddr *sock_address;/*il sockaddr del prossimo client*/
     char *comm_private_key;/*chiave privata commutativa*/
     int max;/*quanti client devono decrittare il messaggio*/
     char *password;/*usata per leggere la chiave privata*/
     struct params_union_intersection_receive_c *par_rec_c;/*contiene i parametri e le stringhe crittate*/
}param_critt_s;


/*struttura usata dai thread della secure_union_intersection per ricevere l'unione degli elementi*/
typedef struct params_clear_r
{
     SSL *ssl_server;/*il server ssl*/
     int len;/*lunghezza stringhe ricezione ed invio*/
     unsigned char ***clear;/*puntatore alla lista di stringhe*/
     unsigned long *n_clear;/*puntatore al numero totale di stringhe ricevute*/
}param_clear_r;


/*struttura usata dai thread della secure_union_intersection del primo host per inviare il risultato a tutti*/
typedef struct params_intersection_s
{
     int len;/*lunghezza stringhe ricezione ed invio*/
     SSL_CTX *ctx_client;/*questo gli serve per inviare*/
     struct sockaddr **sock_addresses;/*sockaddr del prossimo client*/
     int max;/*quanti client hanno gia' crittato i messaggi*/
     unsigned long *n_inter;/*puntatore al numero totale di intersezioni*/
     struct params_union_intersection_receive_c *par_rec_c;/*contiene i parametri e le stringhe crittate*/
}param_intersection_s;


/*struttura usata dai thread della secure_union_intersection per ricevere l'unione degli elementi*/
typedef struct params_intersection_r
{
     SSL *ssl_server;/*il server ssl*/ 
     unsigned long *n_inter;/*puntatore al numero totale di intersezioni*/
}param_intersection_r;


/******************************************************************************/
/*                                                                            */
/*                                PROTOTIPI                                   */
/*                                                                            */
/******************************************************************************/


/*funzioni generali*/
int inet_aton(const char *cp, struct in_addr *inp);
int gethostname(char *name, size_t len);
char **ip_sort(char *ip_file, int *hosts);
psockaddr_in *create_sockaddresses(char **ip_addresses, int hosts_num, char *my_name, char *my_ip, int *pos);
void prepare_sockaddr(struct sockaddr_in *sock_address, int port);
void connect_alarm(int signo);
int connect_timeout(int sockfd, struct sockaddr *sock_address, int nsec);
unsigned long generate_random_long();
int generate_random_int();


/*secure sum*/
unsigned long calculate_value(unsigned long tot_value, unsigned long value);
int secure_sum_send(SSL_CTX* ctx_client, struct sockaddr *sock_address, unsigned long value);
int secure_sum_receive(int rcv_sockfd, SSL* ssl, struct sockaddr *sock_address, unsigned long *value);
unsigned long secure_sum(char **ip_addresses, int hosts_num, unsigned long value);


/*secure union-intersection*/
unsigned char **compare_buffer(unsigned char **buffs, unsigned long len, int size);
unsigned long check_intersection(unsigned char **buffs, unsigned long len, int size, int hosts_num);
void permute_buffer(unsigned char **buffs, unsigned long max);
void *thread_secure_union_intersection_send_rsa(void *ptr);
void *thread_secure_union_intersection_receive_rsa(void *ptr);
int secure_union_send_data(SSL_CTX* ctx, struct sockaddr *sock_address,unsigned char **buf, unsigned long max, int len,
			   char *num, int num_len, int type);
void *thread_secure_union_intersection_c(void *ptr);
void *thread_secure_union_d(void *ptr);
int secure_union_send_critt(SSL_CTX* ctx, struct sockaddr *sock_address, unsigned char **critt, unsigned long n_critt, int len);
void *thread_secure_union_intersection_receive_critt(void *ptr);
void *thread_secure_union_send_critt1(void *ptr);
void *thread_secure_intersection_send_result(void *ptr);
void *thread_secure_intersection_receive_result(void *ptr);
void *thread_secure_union_receive_clear(void *ptr);
unsigned char **secure_union_intersection(char **ip_addresses, int hosts_num, char **sets, unsigned long *total, int protocol);
unsigned char **secure_union(char **ip_addresses, int hosts_num, char **sets, unsigned long *total);
unsigned long secure_intersection(char **ip_addresses, int hosts_num, char **sets);


/*secure scalar product*/
unsigned long pow(int a, int b);
unsigned long *generate_random_vector(int n);
unsigned long **generate_random_matrix(int n, int m);
unsigned long *generate_empty_vector(int n);
unsigned long **generate_empty_matrix(int n, int m);
unsigned long **traspose_matrix(unsigned long **ma, int n, int m);
unsigned long *mult_matrix_vector(unsigned long **ma, unsigned long *v, int n, int m);
unsigned long mult_vector_vector(unsigned long *v1, unsigned long *v2, int n);
unsigned long *sum_vector_vector(unsigned long *v1, unsigned long *v2, int n);
int secure_scalar_send_matrix(SSL_CTX *ctx_client, struct sockaddr *sock_address, unsigned long **matrix, int m, int n);
int secure_scalar_receive_matrix(int rcv_sockfd, SSL* ssl, struct sockaddr *sock_address, unsigned long **matrix, int m, int n);
int secure_scalar_send_vector(SSL_CTX *ctx_client, struct sockaddr *sock_address, unsigned long *vector, int n);
int secure_scalar_receive_vector(int rcv_sockfd, SSL* ssl, struct sockaddr *sock_address, unsigned long *vector, int n);
int secure_scalar_send_scalar(SSL_CTX *ctx_client, struct sockaddr *sock_address, unsigned long scalar);
int secure_scalar_receive_scalar(int rcv_sockfd, SSL* ssl, struct sockaddr *sock_address, unsigned long *scalar);
unsigned long secure_scalar_product(char **ip_addresses, int hosts_num, char *set);


/******************************************************************************/
/*                                                                            */
/*                              FUNZIONI GENERALI                             */
/*                                                                            */
/******************************************************************************/


/*ordina una lista di indirizzi ip scritti sul file "ip_file" e restituisce
  un array ordinato con gli ip e copia in "hosts" il numero totale di indirizzi*/
char **ip_sort(char *ip_file, int *hosts)
{
     FILE *fp; /*il file degli indirizzi ip*/
     char ip_address[MAX_IP_LENGTH];/*temporaneo per l'ip*/
     int hosts_num = 0;/*numero di ip presenti*/
     struct sockaddr_in new_addr, old_addr;/*per comparare gli indirizzi ip*/
     char **ip_addresses = NULL, **tmp_addresses = NULL;/*puntatori agli indirizzi ip*/
     char func_name[] = "ip_sort";/*il nome della funzione*/

     if((fp = fopen(ip_file, "r")) ==  NULL)
     {
	  printf("[%s]: ", func_name);
	  perror("fopen");
	  exit(errno);
     }
     while((fgets(ip_address, MAX_IP_LENGTH, fp)) != NULL)
     {
	  char *new_address;
	  hosts_num++;
	  ip_address[strlen(ip_address) -1] = '\0'; /*tolgo l'a capo*/
	  if(inet_aton(ip_address, &(new_addr.sin_addr)) == 0) /*copio l'ind. e controllo se e' valido*/
	  {
	       printf("Parse error in line %d of file %s\n", hosts_num, ip_file);
	       return NULL; /*errore: formato ip non valido*/
	  }
	  if((new_address = malloc(strlen(ip_address) * sizeof(char))) == NULL)
	  {
	       printf("[%s]: ", func_name);
	       perror("malloc");
	       exit(errno);
	  }
	  strcpy(new_address, ip_address);
	  if((ip_addresses = malloc((hosts_num) * sizeof(char *))) == NULL)/*alloco i puntatori*/
	  {
	       printf("[%s]: ", func_name);
	       perror("malloc");
	       exit(errno);
	  }
	  if(hosts_num == 1)/*se e' il primo lo metto in cima*/
	  {
	       ip_addresses[0] = new_address;
	  }
	  else
	  {
	       int i, j;
	       for(i = 0; i < hosts_num - 1; i++)
	       {
		    if(inet_aton(tmp_addresses[i], &(old_addr.sin_addr)) == 0)/*copio l'indirizzo vecchio*/
		    {
			 return NULL; /*non dovrebbe accadere*/
		    }
		    if(ntohl(new_addr.sin_addr.s_addr) < ntohl(old_addr.sin_addr.s_addr))/*indirizzo piu' basso*/
		    {
			 ip_addresses[i] = new_address;
			 for(j = i + 1; j < hosts_num; j++)/*sposto i valori*/
			 {
			      ip_addresses[j] = tmp_addresses[j - 1];
			 }
			 break;
		    }
		    else/*indirizzo piu' alto*/
		    {
			 ip_addresses[i] = tmp_addresses[i];
		    }
	       }
	       if(i == hosts_num - 1)/*se e' il piu' alto, in fondo*/
	       {
		    ip_addresses[hosts_num - 1] = new_address;
	       }
	  }
	  free(tmp_addresses);/*libero i vecchi puntatori*/
	  tmp_addresses = ip_addresses;/*aggiorno tmp*/

     }/* while(fgets */
     fclose(fp);
     *hosts = hosts_num;/*copio il numero degli host*/
     return ip_addresses;
}


/*crea e ritorna un array di strutture sock_address a partire dalla lista ordinata di ip 
  "ip_addresses", contenente "hosts_num" ip, scrivendo in "my_ip" l'indirizzo ip di "my_name"
  ed in "pos" la sua posizione all'interno dalla lista*/
psockaddr_in *create_sockaddresses(char **ip_addresses, int hosts_num, 
				   char *my_name, char *my_ip, int *pos)
{
     psockaddr_in *sock_addresses = NULL;/*ritornato alla fine*/
     char other_ip[MAX_IP_LENGTH]; /*indirizzo dell'altro client*/
     struct hostent *my_hostent; /*contiene i dati del client*/
     int i;/*per scorrere gli array*/
     char func_name[] = "create_sockaddresses";/*il nome della funzione*/

     /*creo le strutture*/
     if((sock_addresses = malloc((hosts_num) * sizeof(struct sock_addr_in *))) == NULL)/*alloco i puntatori*/
     {
	  printf("[%s]: ", func_name);
	  perror("malloc");
	  exit(errno);
     }
     for(i = 0; i < hosts_num; i++)
     {
	  psockaddr_in new_sock; /*puntatore temp alla nuova struttura*/
	  if((new_sock = malloc(sizeof(struct sockaddr_in))) == NULL) /*alloco il sock_addr*/
	  {
	       printf("[%s]: ", func_name);
	       perror("malloc");
	       exit(errno);
	  }
	  if(inet_aton(ip_addresses[i], &(new_sock->sin_addr)) == 0) /*copio l'ind. e controllo se e' valido*/
	  {
	       printf("Address [%d] of the sorted list is not a valid ip address\n", i);/*non dovrebbe accadere*/
	       return NULL;
	  }
	  sock_addresses[i] = new_sock;
     }
     /*adesso le strutture sono indicizzate in base all'ip*/
     gethostname(my_name, NAME_LENGTH);	
     if((my_hostent = gethostbyname(my_name)) == NULL)
     {       
	  printf("[%s]: ", func_name);
	  perror("gethostbyname");
	  exit(errno);
     }

     /*controllo qual e' la mia posizione*/
     strcpy(my_ip, inet_ntoa(*((struct in_addr *)my_hostent->h_addr)));
     for(i = 0; i < hosts_num; i++)
     {
	  strcpy(other_ip, inet_ntoa(sock_addresses[i]->sin_addr));
	  if(strcmp(my_ip, other_ip) == 0)
	  {
	       *pos = i;
	       break;	
	  }
     }
     if(i == hosts_num)/*il mio ip non esiste*/
     {
	  printf("Your ip address %s has not been found in the ip list.\n", my_ip);
	  exit(1);
     }

     return sock_addresses;
}


/*prepara la struttura "sockaddr" per la ricezione" sulla porta "port"*/
void prepare_sockaddr(struct sockaddr_in *sock_address, int port)
{
     sock_address->sin_family = AF_INET;
     sock_address->sin_port = htons(port);
     memset(&(sock_address->sin_zero), '\0', 8);
}


/*per uscire dalla connect quando arriva il signal alarm*/
void connect_alarm(int signo)
{
     return;
}


/*esegue la connect sul socket "sockfd", utilizzando la stuttura
 "sock_address" aspettando al massimo "nsec" secondi*/
int connect_timeout(int sockfd, struct sockaddr *sock_address, int nsec)
{
     int n;/*per la connect*/
     sighandler_t *sigfunc;/*per la signal*/

     sigfunc = signal(SIGALRM, connect_alarm);
     if(alarm(nsec) != 0)
     {
	  printf("alarm already set\n");
     }
     do
     {
	  n = connect(sockfd, sock_address, sizeof(struct sockaddr));
     }
     while(errno == ECONNREFUSED && n < 0); /*finche' il server non ascolta sto in loop*/
     if(errno == EINTR) errno =ETIMEDOUT;
     alarm(0);
     signal(SIGALRM, sigfunc);
      
     return n;
}


/*ritorna un unsigned long generato in maniera random*/
unsigned long generate_random_long()
{
     unsigned long rand_value = 0;/*il valore random*/
     unsigned char buf[5];/*il buffer in cui inserire i 4 byte random:*/
     int i, k = 0;/*per scorrere il buffer*/
     char func_name[] = "generate_random_long";/*il nome della funzione*/

     if(RAND_bytes(buf, 4) != 1)/*genero 4 bytes in maniera random*/
     {
	  int err = ERR_get_error();
	  ERR_load_crypto_strings();
	  printf("[%s]: RAND_bytes: %s\n", func_name, ERR_error_string(err, NULL));
	  ERR_free_strings();
	  exit(1);
     }
     buf[4] = '\0';
     for(i = 0; i < 4; i++)/*converto i bit in notazione decimale*/
     {
	  int j = 0;
	  for(j = 0; j < 8; j++, buf[i]>>=1)
	  {
	       if(buf[i] & 01)
	       {
		    rand_value += pow(2, k);
	       }
	       k++;
	  }
     }

     return rand_value;
}


/*ritorna un unsigned int generato in maniera random*/
int generate_random_int()
{
     int rand_value = 0;/*il valore random*/
     unsigned char buf[3];/*il buffer in cui inserire i 2 byte random*/
     int i, k = 0;/*per scorrere il buffer*/
     char func_name[] = "generate_random_int";/*il nome della funzione*/

     if(RAND_bytes(buf, 2) != 1)/*genero 2 bytes in maniera random*/
     {
	  int err = ERR_get_error();
	  ERR_load_crypto_strings();
	  printf("[%s]: RAND_bytes: %s\n", func_name, ERR_error_string(err, NULL));
	  ERR_free_strings();
	  exit(1);
     }
     buf[2] = '\0';
     
     for(i = 0; i < 2; i++)/*converto i bit in notazione decimale*/
     {
	  int j = 0;
	  for(j = 0; j < 8; j++, buf[i]>>=1)
	  {
	       if(buf[i] & 01)
	       {
		    rand_value += pow(2, k);
	       }
	       k++;
	  }
     }
 
     return rand_value;
}


/******************************************************************************/
/*                                                                            */
/*                                  SECURE SUM                                */
/*                                                                            */
/******************************************************************************/


/*ritorna il valore per la secure sum come somma del risultato parziale 
  "tot_value" e del valore "value" modulo MOD*/
unsigned long calculate_value(unsigned long tot_value, unsigned long value)
{
     return ((tot_value + value) % MOD);
}


/*invia il valore "value" della secure sum all'host presente in "sock_address"
  sul fd "snd_sockfd"; ritorna 0 in caso di successo altrimenti -1*/
int secure_sum_send(SSL_CTX* ctx_client, struct sockaddr *sock_address, 
		    unsigned long value)
{
     SSL *ssl;/*il client SSL*/
     char buf[BUF_LENGTH];/*il buffer per mandare i messaggi */
     int len;/*lunghezza del buffer*/
     int n;/*per la connect*/
     int snd_sockfd;/*il sending socket*/
     char func_name[] = "secure_sum_send";/*il nome della funzione*/

     ssl = SSL_new(ctx_client);
     if((snd_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("socket");
	  exit(errno);
     }
     if((SSL_set_fd(ssl, snd_sockfd)) == 0)
     {
	  print_ssl_error(func_name, "SSL_set_fd");
	  exit(errno);
     }
     if((n = connect_timeout(snd_sockfd, sock_address, SUM_CONNECT_TIMEOUT)) < 0)
     {
	  return  -1;
     }
     if((n = SSL_connect(ssl)) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_CONNECT");
	  return -1;
     }
	
     sprintf(buf, "%lu", value);
     len = strlen(buf);
     if((n = SSL_write(ssl, buf, len)) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_write");
	  return -1;
     }
     shutdown(snd_sockfd, SHUT_RDWR);
     close(snd_sockfd);
     SSL_free(ssl);
     return 0;
}


/*riceve il valore della secure sum dal fd "rcv_sockfd", salvandolo in "value", 
  utilizzando la struct "sock_address"; ritorna 0 in caso di successo altrimenti -1*/
int secure_sum_receive(int rcv_sockfd, SSL* ssl, struct sockaddr *sock_address,
		       unsigned long *value)
{
     fd_set read_fds;/*per la select*/
     struct timeval tv;/*il timeout per la select*/
     int other_sockfd;/*fd socket dell'altro*/
     char buf[BUF_LENGTH];/*il buffer per ricevere i messaggi */
     int byte_rcv;/*quanti byte ricevuti*/
     socklen_t sin_size;/*dimensione della struct*/
     struct linger li; /*per settare il flag RST dei socket in ascolto (per chiuderli senza stare in TIME_WAIT*/
     int n;/*per la accept*/
     char func_name[] = "secure_sum_receive";/*il nome della funzione*/

     li.l_onoff = 1;
     li.l_linger = 0;
     tv.tv_sec = SUM_CONNECT_TIMEOUT;
     tv.tv_usec = 0;
     FD_ZERO(&read_fds);
     FD_SET(rcv_sockfd, &read_fds);
     sin_size = sizeof(struct sockaddr_in);
     if(select(rcv_sockfd + 1, &read_fds, NULL, NULL, &tv) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("select");
	  exit(errno);
     }
     if(FD_ISSET(rcv_sockfd, &read_fds))/*pronto a leggere dati*/
     {
	  if((other_sockfd = accept(rcv_sockfd, sock_address, &sin_size)) == -1)
	  {
	       printf("[%s]: ", func_name);
	       perror("accept");
	       exit(errno);
	  }
	  else
	  {
	       /*quando chiudero' il socket andra' in shutdown senza passare per il  TIME_WAIT*/
	       setsockopt(other_sockfd, SOCK_STREAM, SO_LINGER, (void *)&li, sizeof(struct linger));
	       SSL_set_fd(ssl, other_sockfd);
	       if((n = SSL_accept(ssl)) < 0)
	       {
		    ssl_get_error(ssl, n, func_name, "SSL_accept");
		    return -1;
	       }
	       byte_rcv = SSL_read(ssl, buf, BUF_LENGTH);
	       if(byte_rcv <= 0)
	       {
		    ssl_get_error(ssl, byte_rcv, func_name, "SSL_read");
		    return -1;
	       }
	       buf[byte_rcv] = '\0';
	       sscanf(buf, "%lu", value); /*converte value in un unsigned long*/
	       shutdown(other_sockfd, SHUT_RDWR);
	       close(other_sockfd);
	  }
     }
     else
     {
	  return -1;
     }

     return 0;
}


/*esegue il protocollo della secure sum tra "hosts_num" host tra gli indirizzi ip 
  ordinati presenti in "ip_addresses", utilizzando il valore "value"; ritorna la somma*/
unsigned long secure_sum(char **ip_addresses, int hosts_num, unsigned long value)
{
     psockaddr_in *sock_addresses;/*puntatori alle strutture sock_add*/
     int i;/*per scorrere l'array*/
     int rcv_sockfd;/*fd socket d'ascolto*/
     int pos = 0, next_pos;/*la posizione dell'ip e del successivo*/	
     char my_name[NAME_LENGTH];/*contiene il nome del client*/
     char my_ip[MAX_IP_LENGTH];/*indirizzo ip mio*/
     unsigned long tot_value;/*il valore ricevuto ed inviato */	
     SSL *ssl_server;/*gli oggetti ssl in invio ed ascolto*/
     SSL_CTX *ctx_client, *ctx_server;/*i context ssl lato client e server*/
     /*nome del file con i certificati pubblici in formato testo*/
     char* CAfile = CAFILE;
     /*nome del file con il certificato del client*/
     char cert[MAX_FILENAME_LENGTH] = CERTIFICATE_DIR;
     /*nome del file con la chiave private del client*/
     char secret_key[MAX_FILENAME_LENGTH] = CERTIFICATE_DIR;
     /*per settare il flag RST dei socket in ascolto (per chiuderli senza stare in TIME_WAIT)*/
     struct linger li; 
     char func_name[] = "secure_sum";/*il nome della funzione*/
     
     if(hosts_num < MIN_CLIENTS_NUM)
     {
	  printf("You haven't provided enough ip addresses.\n");
	  exit(1);
     }
     li.l_onoff = 1;
     li.l_linger = 0;
     ssl_init();
     gethostname(my_name, NAME_LENGTH);	
     sock_addresses = create_sockaddresses(ip_addresses, hosts_num, my_name, my_ip, &pos);
     /*adesso comincia il protocollo di comunicazione*/
     next_pos = (pos + 1) % hosts_num;
     /*preparo i nomi dei file*/
     strcat(cert, my_ip);
     strcat(cert, CERTIFICATE_EXT);
     strcat(secret_key, my_ip);
     strcat(secret_key, PRIVATE_EXT);
     /*preparo il contesto ssl*/
     if((ctx_client = ssl_setup_ctx_client(CAfile, cert, secret_key)) == NULL)
     {
	  exit(1);
     }

     if((ctx_server = ssl_setup_ctx_server(CAfile, cert, secret_key)) == NULL)
     {
	  exit(1);
     }
     /*preparo il receiving socket: pos e' utilizzato come receiving socket*/
     if((rcv_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("socket");
	  exit(errno);
     }
     prepare_sockaddr(sock_addresses[pos], SUM_PORT_NUMBER);
     if(bind(rcv_sockfd, (struct sockaddr *)sock_addresses[pos], sizeof(struct sockaddr)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("bind");
	  exit(errno);
     }   
     if(listen(rcv_sockfd, 1) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("listen");
	  exit(errno);
     }
     /*non sta in TIME_WAIT dopo la close*/
     setsockopt(rcv_sockfd, SOCK_STREAM, SO_LINGER, (void *)&li, sizeof(struct linger));
     if(pos)/*il client: ascolta, invia, ascolta*/
     {
	  /*riceve il valore dal precedente*/
	  if((ssl_server = SSL_new(ctx_server)) == NULL)
	  {
	       print_ssl_error(func_name, "SSL_new");
	       exit(1);
	  }
	  if((secure_sum_receive(rcv_sockfd, ssl_server, 
				 (struct sockaddr *)sock_addresses[pos], &tot_value)) != 0)
	  {
	       printf("Timeout: waiting for data from %s\n", 
		      inet_ntoa(sock_addresses[(pos - 1)]->sin_addr));
	       exit(1);
	  }
	  /*SSL_clear(ssl_server);*/
	  SSL_free(ssl_server);
	  tot_value = calculate_value(tot_value, value);
	  /*invia il valore al successivo*/
	  prepare_sockaddr(sock_addresses[next_pos], SUM_PORT_NUMBER);
	  if((secure_sum_send(ctx_client, 
			      (struct sockaddr *)sock_addresses[next_pos], tot_value)) != 0)
	  {
	       printf("Impossible to connect to %s\n", inet_ntoa(sock_addresses[next_pos]->sin_addr));
	       exit(errno);
	  }  
	  if((ssl_server = SSL_new(ctx_server)) == NULL)
	  {
	       print_ssl_error(func_name, "SSL_new");
	       exit(1);
	  }
	  /*riceve il valore finale dal client 0*/	
	  if((secure_sum_receive(rcv_sockfd, ssl_server, 
				 (struct sockaddr *)sock_addresses[pos], &tot_value)) != 0)
	  {
	       printf("Timeout: waiting for final result from %s\n", 
		      inet_ntoa(sock_addresses[0]->sin_addr));
	       exit(1);
	  }
	  /*SSL_clear(ssl_server);*/
	  SSL_free(ssl_server);
     }/*end if(pos)*/
     else /*il client 0: invia, ascolta, invia */
     {
	  unsigned long rand_value = generate_random_long();
	  tot_value = calculate_value(rand_value, value);
	  prepare_sockaddr(sock_addresses[next_pos], SUM_PORT_NUMBER);
	  /*invia il valore al successivo*/		
	  if((secure_sum_send(ctx_client, 
			      (struct sockaddr *)sock_addresses[next_pos], tot_value)) != 0)
	  {
	       printf("Impossible to connect to %s\n", inet_ntoa(sock_addresses[next_pos]->sin_addr));
	       exit(errno);
	  }
	  /*riceve il valore dall'ultimo*/
	  if((ssl_server = SSL_new(ctx_server)) == NULL)
	  {
	       print_ssl_error(func_name, "SSL_new");
	       exit(1);
	  }	
	  if((secure_sum_receive(rcv_sockfd, ssl_server, 
				 (struct sockaddr *)sock_addresses[pos], &tot_value)) != 0)
	  {
	       printf("Timeout: waiting for data from %s\n", 
		      inet_ntoa(sock_addresses[(hosts_num - 1)]->sin_addr));
	       exit(1);
	  }
	  /*SSL_clear(ssl_server);*/
	  SSL_free(ssl_server);
	  /*adesso invia il risultato a tutti*/
	  tot_value = (tot_value - rand_value) % MOD; /*il risultato finale*/
	  for(i = 1; i < hosts_num; i++)
	  {
	       prepare_sockaddr(sock_addresses[i], SUM_PORT_NUMBER);
	       /*invia il risultato*/
	       if((secure_sum_send(ctx_client, (struct sockaddr *)sock_addresses[i], tot_value)) != 0)
	       {
		    printf("Impossible to send data to %s\n", inet_ntoa(sock_addresses[i]->sin_addr));
		    exit(errno);
	       }
	  }/*end for(i*/
     }/*end else*/
     /*dealloco le strutture*/
     for(i = 0; i < hosts_num; i++)
     {
	  free(sock_addresses[i]);
     }
     free(sock_addresses);
     shutdown(rcv_sockfd, SHUT_RDWR);
     close(rcv_sockfd);
     SSL_CTX_free(ctx_client);
     SSL_CTX_free(ctx_server);
     ssl_end();
     return tot_value;
}


/******************************************************************************/
/*                                                                            */
/*                         SECURE UNION-INTERSECTION                          */
/*                                                                            */
/******************************************************************************/


/*controlla nei buffer contenuti in "buffs", tutti di dimensione "size" e di numero "len"
  se ci sono duplicati e in caso affermativo li rimuove; ritorna un buffer senza duplicati
  (NULL come carattere terminatore)*/
unsigned char **compare_buffer(unsigned char **buffs, unsigned long len, int size)
{
     unsigned long i = 0;/*per scorrere le stringhe*/
     unsigned long j;/*per comparare le stringhe*/
     unsigned char **f_buffs;/*ritornato alla fine*/
     unsigned long dup = 0; /*numero di duplicati*/

     while(i < (len-1))
     {
	  if(buffs[i] != NULL)/*se non l'ho cancellato*/
	  {
	       for(j = i+1; j < len; j++)
	       {
		    if(buffs[j] != NULL)
		    {
			 if(memcmp(buffs[i], buffs[j], size) == 0)/*se uguale lo cancello*/
			 {
			      free(buffs[j]);
			      buffs[j] = NULL;
			      dup++;
			 }
		    }
	       } 
	  }
	  i++;
     }
     
     f_buffs = malloc((len-dup) * sizeof(unsigned char *) + 1);/*lo ritorno alla fine*/
     j = 0;
     i = 0;
     while(i < len)
     {
	  if(buffs[i] != NULL)
	  {
	       f_buffs[j] = buffs[i];
	       j++;
	  }
	  i++;
     }
     f_buffs[len-dup] = NULL;
     free(buffs);

     return f_buffs;
}


/*controlla nei buffer contenuti in "buffs", tutti di dimensione "size" e di numero "len"
  se ci sono duplicati di numero "hosts_num"; ritorna il numero di duplicati*/
unsigned long check_intersection(unsigned char **buffs, unsigned long len, 
				 int size, int hosts_num)
{
     unsigned long i = 0;/*per scorrere le stringhe*/
     unsigned long j;/*per comparare le stringhe*/
     unsigned long n_inter = 0; /*numero di intersezioni*/
     while(i < (len-1))
     {
	  int num_dup = 0;/*quanti duplicati ci sono*/
	  if(buffs[i] != NULL)
	  {
	       num_dup = 1;/*l'elemento che controllo*/
	       for(j = i+1; j < len; j++)
	       {
		    if(buffs[j] != NULL)
		    {
			 if(memcmp(buffs[i], buffs[j], size) == 0)/*se uguale lo cancello*/
			 {
			      free(buffs[j]);
			      buffs[j] = NULL;
			      num_dup++;/*aumento il numero dei duplicati*/
			 }
		    }
	       }
	  }
	  i++;
	  if(num_dup == hosts_num)
	  {
	       n_inter++;
	  }
     }

     return n_inter;
}


/*permuta l'ordine delle stringhe contenute in "buffs" di numero "max"*/
void permute_buffer(unsigned char **buffs, unsigned long max)
{
     unsigned char **perm = malloc(max * sizeof(unsigned char *));/*le stringhe permutate*/
     /*array di numeri random*/
     unsigned long *rand_nums = malloc(max * sizeof(unsigned long));
     unsigned long i;/*per scorrere le stringhe*/

     srand(time(NULL));
     for(i = 0; i < max; i++)
     {
	  unsigned long j = 0;/*per scorrere i numeri gia' creati*/
	  unsigned long rand_num = generate_random_long() % max;
	  while(j < i)/*guardo se l'ho gia' creato uguale*/
	  {
	       if(rand_nums[j] == rand_num)/*sono arrivato alla fine: genero un altro random*/
	       {
		    j = 0;
		    rand_num = generate_random_long() % max;
	       }
	       else
	       {
		    j++;
	       }
	  }
	  rand_nums[i] = rand_num;
     }
     for(i = 0; i < max; i ++)
     {
	  perm[i] = buffs[rand_nums[i]];
     }
     /*cambio l'ordine di buffs*/
     for(i = 0; i < max; i++)
     {
	  buffs[i] = perm[i];
     }
     free(rand_nums);
     free(perm);
}

	
/*thread utilizzato per inviare i valori p e q della chiave RSA commutativa: parametri presenti in "ptr"*/
void *thread_secure_union_intersection_send_rsa(void *ptr)
{
     int n; /*per la connect*/	
     char *buff; /*dove scrivo p e q in esadecimale*/
     int len; /*lunghezza del buffer*/
     int snd_sockfd; /*il sending socket*/
     SSL *ssl; /*il contesto ssl*/
     struct params_rsa_s *par; /*i parametri passati tramite ptr*/
     char c_type[MSG_LENGTH]; /*il tipo del messaggio inviato*/
     char func_name[] = "thread_secure_union_intersection_send_rsa";/*il nome della funzione*/

     par = (struct params_rsa_s*)(ptr);
     ssl = SSL_new(par->ctx_client);
     if((snd_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("socket");
	  exit(errno);
     }
     sprintf(c_type,"%d", SEND_RSA);
     if((SSL_set_fd(ssl, snd_sockfd)) == 0)
     {
	  print_ssl_error(func_name, "SSL_set_fd");
	  exit(errno);
     }
     if((n = connect_timeout(snd_sockfd, par->sock_address, UNION_CONNECT_TIMEOUT)) < 0)
     {
	  printf("[%s]: ", func_name);
	  perror("connect");
	  exit(errno);
     }

     if((n = SSL_connect(ssl)) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_connect");
	  exit(errno);	
     }
     /*invio la richiesta*/
     if((n = SSL_write(ssl, c_type, strlen(c_type))) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_write");
	  exit(errno);
     }
     /*converto il bignum*/
     buff = BN_bn2hex(par->p);
     len = strlen(buff);
     /*invio P*/
     if((n = SSL_write(ssl, buff, len)) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_write");
	  exit(errno);
     }
     free(buff);
     buff = BN_bn2hex(par->q);
     len = strlen(buff);
     /*invio Q*/
     if((n = SSL_write(ssl, buff, len)) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_write");
	  exit(errno);
     }
     shutdown(snd_sockfd, SHUT_RDWR);
     close(snd_sockfd);
     SSL_free(ssl);
     free(buff);

     return NULL;
}


/*thread che riceve i valori p e q per creare la chiave RSA commutativa: parametri presenti in "ptr"*/
void *thread_secure_union_intersection_receive_rsa(void *ptr)
{
     struct params_rsa_r *par;/*i parametri passati tramite ptr*/
     char esa[BIGNUM_BUF_LENGTH];/*contiene la rapp esadecimale di p e q*/
     int byte_rcv;/*quanti byte ricevuti*/
     char func_name[] = "thread_secure_union_intersection_receive_rsa";/*il nome della funzione*/

     par = (struct params_rsa_r*)(ptr);
     /*ricevo P*/
     byte_rcv = SSL_read(par->ssl_server, esa, BIGNUM_BUF_LENGTH);
     if(byte_rcv <= 0)
     {
	  ssl_get_error(par->ssl_server, byte_rcv, func_name, "SSL_read");
	  exit(errno);
     }
     esa[byte_rcv] = '\0';
     BN_hex2bn(par->p, esa);
     /*ricevo q*/
     byte_rcv = SSL_read(par->ssl_server, esa, BIGNUM_BUF_LENGTH);
     if(byte_rcv <= 0)
     {
	  ssl_get_error(par->ssl_server, byte_rcv, func_name, "SSL_read");
	  exit(errno);
     }
     esa[byte_rcv] = '\0';
     BN_hex2bn(par->q, esa);
     SSL_free(par->ssl_server);

     return NULL;
}


/*funzione utlizzata per inviare "max" dati di lunghezza "len" presenti in "buf" utilizzando 
  il contesto "ctx" sul sock "sock_address"; "num" contiene il numero di host che hanno
  (de)crittato i messaggi; "type" specifica il tipo di richiesta*/
int secure_union_send_data(SSL_CTX* ctx, struct sockaddr *sock_address, 
			   unsigned char **buf, unsigned long max, int len, 
			   char *num, int num_len, int type)
{
     int n;/*per la connect*/	
     int j;/*per scorrere i messaggi*/
     char n_max[MSG_LENGTH];/*quanti messaggi invio*/
     SSL* ssl; /*il contesto ssl*/
     int snd_sockfd; /*il sending socket*/
     char c_type[MSG_LENGTH]; /*il tipo di messaggio inviato*/
     char func_name[] = "secure_union_send_data";/*il nome della funzione*/

     sprintf(c_type,"%d", type);
     permute_buffer(buf, max);
     if((snd_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("socket");
	  exit(errno);
     }
     ssl = SSL_new(ctx);
     sprintf(n_max,"%lu", max);
     if((SSL_set_fd(ssl, snd_sockfd)) == 0)
     {
	  print_ssl_error(func_name, "SSL_set_fd");
	  exit(errno);
     }
     if((n = connect_timeout(snd_sockfd, sock_address, UNION_CONNECT_TIMEOUT)) < 0)
     {
	  printf("[%s]: ", func_name);
	  perror("connect");
	  exit(errno);
     }
     if((n = SSL_connect(ssl)) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_connect");
	  exit(errno);
     }
     /*invio la richiesta*/
     if((n = SSL_write(ssl, c_type, strlen(c_type))) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_write");
	  exit(errno);
     }
     /*quanti messaggi*/
     if((n = SSL_write(ssl, n_max, strlen(n_max))) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_write");
	  exit(errno);
     }
     /*quanti l'hanno crittato  */
     if((n = SSL_write(ssl, num, num_len)) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_write");
	  return -1;
     }
     for(j = 0; j < max; j++)
     {
	  if((n = SSL_write(ssl, buf[j], len)) <= 0)
	  {
	       ssl_get_error(ssl, n, func_name, "SSL_write");
	       exit(errno);
	  }
     }
     shutdown(snd_sockfd, SHUT_RDWR);
     close(snd_sockfd);
     SSL_free(ssl);

     return 0;
}


/*thread deputato alla ricezione e crittazione dei messaggi: parametri presenti in "ptr"*/
void *thread_secure_union_intersection_c(void *ptr)
{
     struct params_union_intersection_c *par;/*contiene i parametri passati tramite ptr*/
     unsigned char **buff;/*per la ricezione*/
     unsigned char **mesg;/*per l'invio*/
     int byte_rcv; /*quanti byte ha letto*/
     char num[MSG_LENGTH];/*quanto volte i messaggi sono stati crittati: char*/
     int i; /*quante volte i messaggi sono stati crittati: int*/
     char tot[MSG_LENGTH];/*quanti messaggi mi sta inviando: char*/
     unsigned long max;/*quanti messaggi mi sta inviando: long*/
     unsigned long j;/*per leggere tutti i messaggi*/
     char func_name[] = "thread_secure_union_intersection_c";/*il nome della funzione*/

     par = (struct params_union_intersection_c*)(ptr);
     if(par->other_sockfd > 0)/*se non e' la prima volta*/
     {
	  /*quanti messaggi mi invia*/
	  byte_rcv = SSL_read(par->ssl_server, tot, MSG_LENGTH);
	  if(byte_rcv <= 0)
	  {
	       ssl_get_error(par->ssl_server, byte_rcv, func_name, "SSL_read");
	       exit(1);
	  }
	  tot[byte_rcv] = '\0';
	  sscanf(tot, "%lu", &max);
	  buff = malloc(max * sizeof(unsigned char*)); /*mi alloco i puntatori*/
	  mesg = malloc(max * sizeof(unsigned char*)); /*mi alloco i puntatori*/
	  /*in quanti li hanno crittati*/
	  byte_rcv = SSL_read(par->ssl_server, num, MSG_LENGTH);
	  if(byte_rcv <= 0)
	  {
	       ssl_get_error(par->ssl_server, byte_rcv, func_name, "SSL_read");
	       exit(errno);
	  }
	  num[byte_rcv] = '\0';
	  i = atoi(num);
	  i++;
	  sprintf(num, "%d",i);
	  for(j = 0; j < max; j ++)
	  {	
	       buff[j] = malloc((par->len) * sizeof(unsigned char)); /*mi alloco le stringhe che ricevo*/
	       /*il messaggio crittato*/
	       byte_rcv = SSL_read(par->ssl_server, buff[j], par->len);
	       if(byte_rcv <= 0)
	       {
		    ssl_get_error(par->ssl_server, byte_rcv, func_name, "SSL_read");
		    exit(errno);
	       }
	  }
	  /*adesso che ho ricevuto i messaggi li critto*/
	  for(j = 0; j < max; j ++)
	  {
	       mesg[j] = rsa_encrypt2(par->comm_public_key, buff[j], par->len);
	  }
	  if(i < par->max)/*l'ultima volta non invia al successivo*/
	  {
	       if(par->ctx_client != NULL)
	       {
		    if(secure_union_send_data(par->ctx_client, par->sock_address, mesg,
					      max, par->len, num, strlen(num), ENCRYPT) < 0)
		    {
			 printf("Timeout: sending data to %s\n", inet_ntoa(((psockaddr_in)(par->sock_address))->sin_addr));
			 exit(errno);
		    }
	       }
	       for(j = 0; j < max; j ++)
	       {
		    free(buff[j]);
		    free(mesg[j]);
	       }
	       free(buff);
	       free(mesg);
	  }
	  else/*ritorno le stringhe crittate da tutti al client 0*/
	  {
	       if(par->server_sock_address == NULL)/*il client 0 si copia le stringhe*/
	       {
		    unsigned char **critt = malloc(max * sizeof(unsigned char *) + 1);/*le termina con NULL*/
		    unsigned long j;
		    for(j = 0; j < max; j++)
		    { 
			 critt[j] = mesg[j];
		    }
		    critt[max] = NULL;
		    pthread_mutex_lock(&mutex_critt_params);/*aggiorno la struttura*/
		    par->par_rec_c->n_critt += max;/*aumento numero crittogrammi*/
		    par->par_rec_c->critt[par->par_rec_c->tot] = critt;/*aggiungo i messaggi crittati*/
		    par->par_rec_c->tot += 1;/*aumento numero totale*/
		    if(par->par_rec_c->tot == par->par_rec_c->max)
		    {
			 pthread_cond_broadcast(&cond_critt_params);
		    }
		    pthread_mutex_unlock(&mutex_critt_params);
		    for(j = 0; j < max; j ++)
		    {
			 free(buff[j]);/*le stringhe in mesg non lo dealloco*/
		    }
		    free(buff);
		    free(mesg);
	       }
	       else/*sono il client 1, 2...*/
	       {	
		    if(secure_union_send_critt(par->ctx_client, par->server_sock_address, mesg, max, par->len) < 0)
		    {
			 printf("Error on secure_union_send_critt\n");
			 return NULL;
		    }
		    for(j = 0; j < max; j ++)
		    {
			 free(buff[j]);
			 free(mesg[j]);
		    }
		    free(buff);
		    free(mesg);
	       }/*else sono il client 1*/
	  }/*else ritorno le stringhe*/
     }
     else/*invio la prima volta i miei messagi*/
     {
	  /*crittare tutti i messaggi ed inviarli*/
	  unsigned char **critt = malloc(par->n_clear *sizeof(unsigned char *));
	  for(j = 0; j < par->n_clear; j++)
	  {
	       critt[j] = rsa_encrypt1(par->comm_public_key, par->clear[j], strlen(par->clear[j]));
	  }
	  sprintf(num, "1");/*e' la prima volta*/
	  if(secure_union_send_data(par->ctx_client, par->sock_address, critt, 
				    par->n_clear, par->len, num, strlen(num), ENCRYPT) < 0)
	  {
	       printf("Timeout: sending data to %s\n", inet_ntoa(((psockaddr_in)(par->sock_address))->sin_addr));
	       exit(errno);
	  }
	  for(j = 0; j < par->n_clear; j++)
	  {
	       free(par->clear[j]);
	       free(critt[j]);
	  }
	  free(par->clear);
	  free(critt);
     }
     SSL_free(par->ssl_server);
     free(par->comm_public_key);
     free(par);

     return NULL;
}


/*thread deputato alla ricezione decrittazione dei messaggi: parametri presenti in "ptr"*/
void *thread_secure_union_d(void *ptr)
{
     struct params_union_d *par;
     unsigned char **buff;/*per la ricezione*/
     unsigned char **mesg;/*per l'invio*/
     int byte_rcv;/*quanti byte ricevuti*/
     char num[MSG_LENGTH];/*quanto volte i messaggi sono stati crittati: char*/
     int i; /*quante volte i messaggi sono stati crittati: int*/
     char tot[MSG_LENGTH];/*quanti messaggi mi sta inviando: char*/
     unsigned long max;/*quanti messaggi mi sta inviando: long*/
     unsigned long j;/*per leggere tutti i messaggi*/
     char func_name[] = "thread_secure_union_d";/*il nome della funzione*/

     par = (struct params_union_d*)(ptr);
     /*quanti messaggi mi invia*/
     byte_rcv = SSL_read(par->ssl_server, tot, MSG_LENGTH);
     if(byte_rcv <= 0)
     {
	  ssl_get_error(par->ssl_server, byte_rcv, func_name, "SSL_read");
	  exit(errno);
     }
     tot[byte_rcv] = '\0';
     sscanf(tot, "%lu", &max);
     buff = malloc(max * sizeof(unsigned char*)); /*mi alloco i puntatori*/
     mesg = malloc(max * sizeof(unsigned char*)); /*mi alloco i puntatori*/
     /*in quanti li hanno decrittati*/
     byte_rcv = SSL_read(par->ssl_server, num, MSG_LENGTH);
     if(byte_rcv <= 0)
     {
	  ssl_get_error(par->ssl_server, byte_rcv, func_name, "SSL_read");
	  exit(errno);
     }
     num[byte_rcv] = '\0';
     i = atoi(num);
     i++;
     sprintf(num, "%d",i);
     for(j = 0; j < max; j ++)
     {	
	  buff[j] = malloc((par->len) * sizeof(unsigned char)); /*mi alloco le stringhe che ricevo*/
	  mesg[j] = malloc((par->len) * sizeof(unsigned char)); /*mi alloco le stringhe che invio*/
	  /*il messaggio crittato*/
	  byte_rcv = SSL_read(par->ssl_server, buff[j], par->len);
	  if(byte_rcv <= 0)
	  {
	       ssl_get_error(par->ssl_server, byte_rcv, func_name, "SSL_read");
	       exit(errno);
	  }
     }
     /*adesso che ho ricevuto i messaggi li decritto*/
     for(j = 0; j < max; j ++)
     {
	  if(rsa_decrypt(par->comm_private_key, buff[j], mesg[j], par->len, par->password) < 0)
	  {
	       printf("Rsa decrypt error\n");
	       exit(errno);
	  }
     }
     if(i < par->max)/*non sono state decrittate da tutti*/
     {
	  if(secure_union_send_data(par->ctx_client, par->sock_address, mesg,
				    max, par->len, num, strlen(num), DECRYPT) < 0)
	  {
	       printf("Timeout: sending data to %s\n", inet_ntoa(((psockaddr_in)(par->sock_address))->sin_addr));
	       exit(errno);
	  }
	  for(j = 0; j < max; j ++)
	  {
	       free(buff[j]);
	       free(mesg[j]);
	  }
	  free(buff);
	  free(mesg);
     }
     else /*ritorno le stringhe chiare a tutti*/
     {
	  unsigned long i;/*per scorrere le stringhe*/
	  char c_type[MSG_LENGTH];/*il tipo di richiesta*/
	  char n_max[MSG_LENGTH];/*quanti messaggi invio*/
	  for(i = 0; i < (par->max - 1); i++)
	  {
	       SSL *ssl;/*il client ssl*/
	       int snd_sockfd;/*il sending socket*/
	       int n;/*per la connect*/
	       ssl = SSL_new(par->ctx_client);
	       if((snd_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	       {
		    printf("[%s]: ", func_name);
		    perror("socket");
		    exit(errno);
	       }
	       sprintf(c_type,"%d", SEND_CLEAR);
	       if((SSL_set_fd(ssl, snd_sockfd)) == 0)
	       {
		    print_ssl_error(func_name, "SSL_set_fd");
		    exit(errno);
	       }
	       if((n = connect_timeout(snd_sockfd, par->server_sock_address[i], 
				       UNION_CONNECT_TIMEOUT)) < 0)
	       {
		    printf("[%s]: ", func_name);
		    exit(errno);
	       }
	       if((n = SSL_connect(ssl)) <= 0)
	       {
		    ssl_get_error(ssl, n, func_name, "SSL_connect");
		    exit(errno);	
	       }
	       /*invio la richiesta*/
	       if((n = SSL_write(ssl, c_type, strlen(c_type))) <= 0)
	       {
		    ssl_get_error(ssl, n, func_name, "SSL_write");
		    exit(errno);
	       }
	       sprintf(n_max, "%lu", max);
	       /*invio il numero di stringhe*/
	       if((n = SSL_write(ssl, n_max, strlen(n_max))) <= 0)
	       {
		    ssl_get_error(ssl, n, func_name, "SSL_write");
		    exit(errno);
	       }
	       /*invio le stringhe*/
	       for(j = 0; j < max; j++)
	       {
		    if((n = SSL_write(ssl, mesg[j], par->len)) <= 0)
		    {
			 ssl_get_error(ssl, n, func_name, "SSL_write");
			 exit(errno);
		    }
	       }
	       shutdown(snd_sockfd, SHUT_RDWR);
	       close(snd_sockfd);
	       SSL_free(ssl);
	  }/*for*/
	  *(par->clear) = mesg;
	  *(par->n_clear) = max;
	  for(j = 0; j < max; j ++)
	  {
	       free(buff[j]);
	  }
	  free(buff);
     }/*else*/
     SSL_free(par->ssl_server);
     free(par->comm_private_key);
     free(par);

     return NULL;
}


/*invia gli elementi "critt" crittati da tutti di numero "n_critt" e lunghezza "len" 
  al server presente in "sock_address" utilizzando "ctx:*/
int secure_union_send_critt(SSL_CTX* ctx, struct sockaddr *sock_address, unsigned char **critt,
			    unsigned long n_critt, int len)
{
     SSL* ssl; /*il contesto SSL*/
     int snd_sockfd; /*il sending socket*/
     char max[MSG_LENGTH];/*quanti messaggi gli invio*/
     unsigned long j;/*per scorrere l'array*/
     int n;/*per la connect*/
     char c_type[MSG_LENGTH];/*il tipo di messaggio*/
     char func_name[] = "secure_union_send_critt";/*il nome della funzione*/


     permute_buffer(critt, n_critt);
     sprintf(c_type,"%d", SEND_CRYPT);
     if((snd_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("socket");
	  exit(errno);
     }
     ssl = SSL_new(ctx);
     sprintf(max,"%lu", n_critt);
     if((SSL_set_fd(ssl, snd_sockfd)) == 0)
     {
	  print_ssl_error(func_name, "SSL_set_fd");
	  exit(errno);
     }
     if((n = connect_timeout(snd_sockfd, sock_address, UNION_CONNECT_TIMEOUT)) < 0)
     {
	  printf("[%s]: ", func_name);
	  exit(errno);
     }
     if((n = SSL_connect(ssl)) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_connect");
	  exit(errno);
     }
     /*invio la richiesta*/
     if((n = SSL_write(ssl, c_type, strlen(c_type))) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_write");
	  exit(errno);
     }
     /*quanti messaggi*/
     if((n = SSL_write(ssl, max, strlen(max))) <= 0)
     {
	  ssl_get_error(ssl, n, func_name, "SSL_write");
	  exit(errno);
     }
     /*loop invio messaggi */
     for(j = 0; j < n_critt; j++)
     {
	  if((n = SSL_write(ssl, critt[j], len)) <= 0)
	  {
	       ssl_get_error(ssl, n, func_name, "SSL_write");
	       exit(errno);
	  }
     }
     shutdown(snd_sockfd, SHUT_RDWR);
     close(snd_sockfd);
     SSL_free(ssl);

     return 0;
}


/*thread (lanciati dal primo host) che riceve le stringhe crittate da tutti (per poi confrontarle):
  parametri presenti in "ptr"*/
void *thread_secure_union_intersection_receive_critt(void *ptr)
{
     struct params_union_intersection_receive_c *par; /*i parametri passati attraverso ptr*/
     int byte_rcv;/*quanti byte ha letto*/
     char tot[MSG_LENGTH];/*quanti messaggi mi sta inviando: char*/
     unsigned long n_critt;/*quanti messaggi mi sta inviando: int*/
     unsigned j; /*per scorrere l'array*/
     unsigned char **critt; /*i crittogrammi letti*/
     SSL *ssl; /*il contesto ssl*/
     char func_name[] = "thread_secure_union_intersection_receive_critt";/*il nome della funzione*/

     par = (struct params_union_intersection_receive_c *)(ptr);
     ssl = par->ssl_server;
     byte_rcv = SSL_read(ssl, tot, MSG_LENGTH);
     if(byte_rcv <= 0)
     {
	  ssl_get_error(ssl, byte_rcv, func_name, "SSL_read");
	  exit(errno);
     }
     tot[byte_rcv] = '\0';
     sscanf(tot, "%lu", &n_critt);
     critt = malloc(n_critt * sizeof(unsigned char *) + 1);/*li termino con NULL, almeno so dove finisce l'array*/
     for(j = 0; j < n_critt; j++)
     {
	  critt[j] = malloc(par->len * sizeof(char)); 
	  byte_rcv = SSL_read(ssl, critt[j], par->len);
	  if(byte_rcv <= 0)
	  {
	       ssl_get_error(ssl, byte_rcv, func_name, "SSL_read");
	       exit(errno);
	  }
     }
     critt[n_critt] = NULL;
     pthread_mutex_lock(&mutex_critt_params);/*aggiorno la struttura*/
     par->n_critt += n_critt;/*aumento numero crittogrammi*/
     par->critt[par->tot] = critt;/*aggiungo i messaggi crittati*/
     par->tot += 1;/*aumento numero invii*/
     if(par->tot == par->max)
     {
          pthread_cond_broadcast(&cond_critt_params);
     }
     pthread_mutex_unlock(&mutex_critt_params);
     SSL_free(ssl);

     return NULL;
}


/*thread (lanciato dal primo host) per inviare la prima volta le stringhe crittate da decrittare:
  parametri presenti in "ptr*/
void *thread_secure_union_send_critt1(void *ptr)
{
     struct params_critt_s *par; /*i parametri ricevuti tramite ptr*/
     unsigned char **critt;/*le stringhe che decritto*/
     unsigned long i;/*per scorrere gli array interni*/
     unsigned long k;/*quanti messaggi ho decrittato. k = n_critt*/
     unsigned long j;/*per scorrere l'array esterno. j = par->max*/
     unsigned long n_critt;/*conterra' il numero delle stringhe diverse*/
     unsigned char **final_critt;/*risultato finale*/
     char num[MSG_LENGTH]; /*quanto volte i messaggi sono stati decrittati: char*/
 
     pthread_mutex_lock(&mutex_critt_params);	
     par = (struct params_critt_s*)(ptr);
     while(par->par_rec_c->tot < par->par_rec_c->max)/*aspetto di averli ricevuti tutti*/
     {
	  pthread_cond_wait(&cond_critt_params, &mutex_critt_params);
     }
     final_critt  = malloc(par->par_rec_c->n_critt *sizeof(unsigned char *));/*le stringhe senza duplicati*/     
     k = 0;
     j = 0;
     i = 0;
     while(j < par->max)/*preparo le stringhe per togliere i duplicati*/
     {
	  if(par->par_rec_c->critt[j][i] != NULL)
	  {
	       final_critt[k] = malloc((par->len) * sizeof(unsigned char));
	       memcpy(final_critt[k], par->par_rec_c->critt[j][i], par->len);
	       k++;
	       i++;
	  }
	  else
	  {
	       j++;
	       i=0;
	  }
     }
     /*tolgo i duplicati*/
     if(par->par_rec_c->n_critt > 0)
     {
	  final_critt = compare_buffer(final_critt, par->par_rec_c->n_critt, par->len);
	  i = 0;
	  while(final_critt[i] != NULL && i < par->par_rec_c->n_critt)
	  {
	       i++;
	  }
	  n_critt = i;
	  critt = malloc((n_critt) * sizeof(unsigned char *)); /*mi alloco le stringhe che invio*/
     }
     else
     {
	  n_critt = 0;
     }
     for(i = 0; i < n_critt; i++)
     {
	  critt[i] = malloc((par->len) * sizeof(unsigned char ));
	  if(rsa_decrypt(par->comm_private_key, final_critt[i], critt[i], par->len, par->password) < 0)
	  {
	       printf("Rsa decrypt error\n");
	       exit(errno);
	  }
     }
     sprintf(num, "1");/*sono il primo a decrittare*/
     if(secure_union_send_data(par->ctx_client, par->sock_address, critt,
			       n_critt, par->len, num, strlen(num), DECRYPT) < 0)
     {
	  printf("Timeout: sending data to %s\n", inet_ntoa(((psockaddr_in)(par->sock_address))->sin_addr));
	  exit(errno);
     }
     for(j = 0; j < n_critt; j++)
     {
	  free(final_critt[j]);
	  free(critt[j]);
     }
     free(final_critt);
     free(critt);
     free(par->comm_private_key);
     free(par);
     pthread_mutex_unlock(&mutex_critt_params);	
     
     return NULL;
}


/*thread (lanciato dal primo host) per ricevere le stringhe crittate da tutti ed inviare il risultato finale:
  parametri presenti in "ptr*/
void *thread_secure_intersection_send_result(void *ptr)
{
     struct params_intersection_s *par; /*i parametri ricevuti tramite ptr*/
     unsigned long i;/*per scorrere gli array interni*/
     unsigned long k;/*quanti messaggi ho decrittato. k = n_critt*/
     unsigned long j;/*per scorrere l'array esterno. j = par->max*/
     unsigned char **final_critt;/*risultato finale*/
     char c_type[MSG_LENGTH];/*quanto volte i messaggi sono stati crittati: char*/
     char c_num[MSG_LENGTH];/*conterra' il risultato finale*/
     char func_name[] = "thread_secure_intersection_send_result";

     pthread_mutex_lock(&mutex_critt_params);
     par = (struct params_intersection_s*)(ptr);
     while(par->par_rec_c->tot < par->par_rec_c->max)/*aspetto di averli ricevuti tutti*/
     {
	  pthread_cond_wait(&cond_critt_params, &mutex_critt_params);
     }
     final_critt  = malloc(par->par_rec_c->n_critt * sizeof(unsigned char *));/*le stringhe senza duplicati*/     
     k = 0;
     j = 0;
     i = 0;
     while(j < par->max)/*preparo le stringhe per contare i duplicati*/
     {
	  if(par->par_rec_c->critt[j][i] != NULL)
	  {
	       final_critt[k] = malloc((par->len) * sizeof(unsigned char));
	       memcpy(final_critt[k], par->par_rec_c->critt[j][i], par->len);
	       k++;
	       i++;
	  }
	  else
	  {
	       j++;
	       i=0;
	  }
     }
     /*conto i duplicati*/
     if(par->par_rec_c->n_critt > 0)
     {
	  *par->n_inter = check_intersection(final_critt, par->par_rec_c->n_critt, 
					     par->len, par->max);
	  sprintf(c_num,"%lu", *par->n_inter);
     }
     else
     {
	  *par->n_inter = 0;
	  sprintf(c_num,"%lu", *par->n_inter);
	  
     }
     /*invio il risultato a tutti*/
     for(i = 1; i < (par->max); i++)
     {
	  SSL *ssl;
	  int snd_sockfd;
	  int n;
	  ssl = SSL_new(par->ctx_client);
	  if((snd_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	  {
	       printf("[%s]: ", func_name);
	       perror("socket");
	       exit(errno);
	  }
	  sprintf(c_type,"%d", SEND_INTERSECTION);
	  if((SSL_set_fd(ssl, snd_sockfd)) == 0)
	  {
	       print_ssl_error(func_name, "SSL_set_fd");
	       exit(errno);
	  }
	  if((n = connect_timeout(snd_sockfd, par->sock_addresses[i], UNION_CONNECT_TIMEOUT)) < 0)
	  {
	       printf("[%s]: ", func_name);
	       exit(errno);
	  }
	  if((n = SSL_connect(ssl)) <= 0)
	  {
	       ssl_get_error(ssl, n, func_name, "SSL_connect");
	       exit(errno);	
	  }
	  if((n = SSL_write(ssl, c_type, strlen(c_type))) <= 0)
	  {
	       ssl_get_error(ssl, n, func_name, "SSL_write");
	       exit(errno);
	  }
	  if((n = SSL_write(ssl, c_num, strlen(c_num))) <= 0)
	  {
	       ssl_get_error(ssl, n, func_name, "SSL_write");
	       exit(errno);
	  }
	  shutdown(snd_sockfd, SHUT_RDWR);
	  close(snd_sockfd);
	  SSL_free(ssl);
     }
     for(j = 0; j < par->par_rec_c->n_critt; j++)
     {
	  if(final_critt[j] != NULL)
	  {
	       free(final_critt[j]);
	  }
     }
     free(final_critt);
     free(par);
     pthread_mutex_unlock(&mutex_critt_params);

     return NULL;
}


/*thread deputato alla ricezione del risultato dell'intersezione: parametri presenti in "ptr"*/
void *thread_secure_intersection_receive_result(void *ptr)
{
     struct params_intersection_r *par; /*i parametri ricevuti tramite ptr*/
     SSL *ssl; /*il contesto ssl*/
     char tot[MSG_LENGTH];/*il risultato finale: char*/
     int byte_rcv;/*quanti byte ha letto*/
     char func_name[] = "thread_secure_intersection_receive_result";/*il nome della funzione*/

     par = (struct params_intersection_r*)(ptr);         
     ssl = par->ssl_server;
     byte_rcv = SSL_read(ssl, tot, MSG_LENGTH);
     if(byte_rcv <= 0)
     {
	  ssl_get_error(ssl, byte_rcv, func_name, "SSL_read");
	  exit(errno);
     }
     tot[byte_rcv] = '\0';
     *par->n_inter = atoi(tot);
     SSL_free(par->ssl_server);

     return NULL;
}


/*thread deputato alla ricezione delle stringhe (finali) in chiaro (cioe' l'unione):
  parametri presenti in "ptr*/
void *thread_secure_union_receive_clear(void *ptr)
{
     struct params_clear_r *par; /*i parametri ricevuti tramite ptr*/
     char tot[MSG_LENGTH];/*quanti messaggi mi sta inviando: char*/
     unsigned long n_clear;/*quanti messaggi mi sta inviando: long*/
     int byte_rcv;/*quanti byte ha letto*/
     SSL *ssl; /*il contesto ssl*/
     unsigned long j; /*per scorrere gli array*/
     unsigned char **clear; /*le stringhe in chiaro*/
     char func_name[] = "thread_secure_union_receive_clear";/*il nome della funzione*/

     par = (struct params_clear_r*)(ptr);
     ssl = par->ssl_server;
     byte_rcv = SSL_read(ssl, tot, MSG_LENGTH);
     if(byte_rcv <= 0)
     {
	  ssl_get_error(ssl, byte_rcv, func_name, "SSL_read");
	  exit(errno);
     }
     tot[byte_rcv] = '\0';
     sscanf(tot, "%lu", &n_clear);
     clear = malloc(n_clear * sizeof(unsigned char *));
     for(j = 0; j < n_clear; j++)
     { 
	  clear[j] = malloc(par->len * sizeof(unsigned char ));
	  byte_rcv = SSL_read(ssl, clear[j], par->len);
	  if(byte_rcv <= 0)
	  {
	       ssl_get_error(ssl, byte_rcv, func_name, "SSL_read");
	       exit(errno);
	  }
     }
     *par->clear = clear;
     *par->n_clear = n_clear;
     SSL_free(par->ssl_server);

     return NULL;
}


/*esegue il protocollo della secure union e della secure interesection 
  a seconda del valore di "protocol" tra "hosts_num" host tra gli indirizzi ip 
  ordinati presenti in "ip_addresses", utilizzando l'array "sets" (terminato da NULL)
  di elementi di dimensione tutti multipli della stessa chiave (terminati da '\0').
  Ritorna l'unione degli insiemi, scrivendo in "total: la dimensione*/
unsigned char **secure_union_intersection(char **ip_addresses, int hosts_num, char **sets, 
					  unsigned long *total, int protocol)
{
     psockaddr_in *sock_addresses; /*puntatori alle strutture sock_add*/
     unsigned long i; /*per scorrere l'array*/
     int j;/*per scorrere i sock*/
     int rcv_sockfd; /*fd socket d'ascolto*/
     int pos = 0, next_pos; /*la posizione dell'ip e del successivo*/	
     char my_name[NAME_LENGTH]; /*contiene il nome del client*/
     char my_ip[MAX_IP_LENGTH]; /*indirizzo ip mio*/	
     /*SSL *ssl_client, *ssl_server;*/ /*gli oggetti ssl in invio ed ascolto*/
     SSL_CTX *ctx_client, *ctx_server; /*i context ssl lato client e server*/
     char* CAfile = CAFILE;/*nome del file con i certificati pubblici in formato testo*/
     char cert[MAX_FILENAME_LENGTH] = CERTIFICATE_DIR;/*nome del file con il certificato del client*/
     char secret_key[MAX_FILENAME_LENGTH] = CERTIFICATE_DIR;/*nome del file con la chiave priv.del client*/
     RSA *common_rsa;/*la chiave RSA contenente il modulo comune*/
     RSA *my_rsa;/*la chiava RSA commutativa dell'host*/
     char comm_public_key[MAX_FILENAME_LENGTH] = CERTIFICATE_DIR;/*chiave pubb. commutativa*/
     char comm_private_key[MAX_FILENAME_LENGTH] = CERTIFICATE_DIR;/*chiave priv. commutativa*/
     char *password = DEFAULT_COMMUTATIVE_PASSWORD; /*password per salvare la chiave commutativa privata*/
     unsigned long n_clear = 0;/*numero stringhe in chiaro*/
     char **clear;/*le stringhe in chiaro*/
     int len_c;/*lunghezza stringhe*/
     int len_s = 0;/*lunghezza stringhe dopo crittatura (uguale per tutti)*/
     BIGNUM *p=NULL, *q=NULL;/*p e q per generare rsa*/
     fd_set read_fds;/*per la select*/
     struct timeval tv;/*il timeout per la select*/
     int other_sockfd;/*fd socket dell'altro*/
     /*char buf[BUF_LENGTH];*/
     socklen_t sin_size;/*dimensione della struct*/
     int first = 1;/*se e la prima volta*/
     int send_first = 0;/*posso inviare*/
     int end = 0;/*ho finito*/
     struct params_union_intersection_receive_c par_rec_c;/*serve al primo host per contenere le stringhe crittate da tutti*/
     unsigned char **final_clear = NULL; /*ritornate alla fine*/
     unsigned long final_n_clear; /*il numero delle stringhe ritornate (copiato in total)*/
     pthread_t ptid[MAX_THREADS_NUMBER]; /*conterra' gli id dei thread creati**/
     int n_thread = 0; /*quanti thread creati*/
     int other_sockfds[MAX_THREADS_NUMBER];/*conterra' i sockfd restituiti dalla connect*/
     int other_sockfds_num = 0;/*numero di sockfd creati dopo la connect*/
     struct linger li; /*per settare il flag RST dei socket in ascolto (per chiuderli senza stare in TIME_WAIT*/
     unsigned long n_inter; /*il risultato dell'intersection*/
     int n_send_crypt = 0; /*solo per la secure intersection:conta il numero di richieste SEND_CRYPT ricevute*/
     int n_encrypt = 0; /*solo per la secure intersection: conta il numero di richieste ENCRYPT ricevute*/
     int RCV_PORT_NUMBER;/*porta d'ascolto*/
     char func_name[] = "secure_union_intersection";/*il nome della funzione*/
     
     if(protocol == SECURE_UNION)
     {
	  RCV_PORT_NUMBER = UNION_PORT_NUMBER;
     }
     else 
     {
	  RCV_PORT_NUMBER = INTER_PORT_NUMBER;
     }
     
     if(hosts_num < MIN_CLIENTS_NUM)
     {
	  printf("You haven't provided enough ip addresses.\n");
	  exit(1);
     }
     li.l_onoff = 1;
     li.l_linger = 0;
     ssl_init();/*inizializzo ssl*/
     gethostname(my_name, NAME_LENGTH);	
     sock_addresses = create_sockaddresses(ip_addresses, hosts_num, my_name, my_ip, &pos);
     /*mi leggo le stringhe passate*/
     clear = sets;
     n_clear = 0;
     if(clear == NULL)
     {
	  n_clear = 0;
	  len_c = DEFAULT_BLOCK_LENGTH;
     }
     else
     {
	  while(clear[n_clear] != NULL)
	  {
	       n_clear++;
	  }
	  if(n_clear > 0)
	  {
	       len_c = strlen(clear[0]);/*devono essere tutte multipli della stessa chiave*/
	  }
	  else
	  {
	       len_c = DEFAULT_BLOCK_LENGTH;
     	}
     }
     /*adesso comincia il protocollo di comunicazione*/
     next_pos = (pos + 1) % hosts_num;
     /*preparo le stringhe per i file*/
     strcat(cert, my_ip);
     strcat(cert, CERTIFICATE_EXT);
     strcat(secret_key, my_ip);
     strcat(secret_key, PRIVATE_EXT);
     strcat(comm_public_key, my_ip);
     strcat(comm_public_key, COMM_PUB_KEY_EXT);
     strcat(comm_private_key, my_ip);
     strcat(comm_private_key, COMM_PRI_KEY_EXT);
     /*creo i contesti client e server*/
     if((ctx_client = ssl_setup_ctx_client(CAfile, cert, secret_key)) == NULL)
     {
	  exit(errno);
     }
     if((ctx_server = ssl_setup_ctx_server(CAfile, cert, secret_key)) == NULL)
     {
	  exit(errno);
     }
     /*preparo il receiving socket*/
     if((rcv_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
     {     
	  printf("[%s]: ", func_name);
	  perror("socket");
	  exit(errno);
     }
     prepare_sockaddr(sock_addresses[pos], RCV_PORT_NUMBER);
     if(bind(rcv_sockfd, (struct sockaddr *)sock_addresses[pos], sizeof(struct sockaddr)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("bind");
	  exit(errno);
     }   
     if(listen(rcv_sockfd, 1) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("listen");
	  exit(errno);
     }
     /*non sta in TIME_WAIT dopo la close*/
     setsockopt(rcv_sockfd, SOCK_STREAM, SO_LINGER, (void *)&li, sizeof(struct linger));
     /*il primo client (lo 0) invia la chiave RSA tramite thread ed invia il primo messaggio
       e mette un thread in attesa che siano arrivati tutti i messagi crittati*/
     if(!pos)
     {
	  /*parametri da passare al primo thread*/
	  struct params_union_intersection_c *par = malloc(sizeof(struct params_union_intersection_c));
	  pthread_t tid;/*il tid del thread*/
	  common_rsa = RSA_gen_key();
	  /*invia i parametri RSA*/
	  for(j = 1; j < hosts_num; j++)
	  {
	       struct params_rsa_s par_rsa;
	       prepare_sockaddr(sock_addresses[j], RCV_PORT_NUMBER);
	       par_rsa.ctx_client = ctx_client;
	       par_rsa.sock_address = (struct sockaddr *)sock_addresses[j];
	       par_rsa.p = common_rsa->p;
	       par_rsa.q = common_rsa->q;
	       pthread_create(&tid, NULL, &thread_secure_union_intersection_send_rsa, &par_rsa);
	       pthread_join(tid, NULL);
	  }
	  my_rsa = RSA_generate_key_commutative(MODULUS_SIZE, common_rsa->p, common_rsa->q);
	  if((write_rsa_keys(comm_public_key, comm_private_key, my_rsa, password)) < 0)
	  {
	       printf("Error while saving commutative keys\n");
	       exit(errno);
	  }
	  free(common_rsa);
	  free(my_rsa);
	  /*adesso invia i suoi messaggi crittandoli*/
	  len_s = calculate_size(comm_public_key, len_c);
	  prepare_sockaddr(sock_addresses[next_pos], RCV_PORT_NUMBER);
	  par->other_sockfd = 0;/*non bisogna ricevere*/;
	  par->ssl_server = NULL;
	  par->len = len_s;
	  par->ctx_client = ctx_client;
	  par->sock_address = (struct sockaddr *)sock_addresses[next_pos];
	  par->comm_public_key = malloc((strlen(comm_public_key) + 1) * sizeof(char));
	  strcpy(par->comm_public_key, comm_public_key);
	  par->clear = malloc(n_clear * sizeof(char *));/*alloco i puntatori*/
	  par->max = len_s;
	  par->n_clear = n_clear;
	  for(i = 0; i < n_clear; i++)
	  {
	       par->clear[i] = malloc((len_s+1) * sizeof(char));
	       strcpy(par->clear[i], clear[i]);
	  }
	  pthread_create(&ptid[n_thread], NULL, &thread_secure_union_intersection_c, par);
	  n_thread++;
	  /*preparo i parametri per contenere le stringhe crittate da tutti*/
	  par_rec_c.max = hosts_num ;/*in quanti devono inviare i crittogrammi, compreso il server*/
	  par_rec_c.critt = malloc(hosts_num * sizeof(unsigned char**));
	  par_rec_c.tot = 0;
	  par_rec_c.n_critt = 0;
	  par_rec_c.len = len_s;
	  if(protocol == SECURE_UNION)/*secure union*/
	  {
	       struct params_critt_s *par2 = malloc(sizeof(struct params_critt_s));
	       /*parametri da passare al secondo
		 aspetta che siano arrivati tutti i messaggi crittati, 
		 toglie i duplicati ed invia al primo host*/
	       par2->ctx_client = ctx_client;
	       prepare_sockaddr(sock_addresses[next_pos], RCV_PORT_NUMBER);
	       par2->sock_address = (struct sockaddr *)sock_addresses[next_pos];
	       par2->comm_private_key = malloc((strlen(comm_private_key) + 1) * sizeof(char));
	       strcpy(par2->comm_private_key, comm_private_key);
	       par2->max = hosts_num;
	       par2->len = len_s;
	       par2->password = password;
	       par2->par_rec_c = &par_rec_c; 
	       pthread_create(&ptid[n_thread], NULL, &thread_secure_union_send_critt1, par2);
	       n_thread++;
	  }
	  else/*secure intersection*/
	  {
	       struct params_intersection_s *par2 = malloc(sizeof(struct params_intersection_s));/*parametri da passare al secondo*/
	       par2->ctx_client = ctx_client;
	       for(i = 1; i < hosts_num; i++)
	       {
		    prepare_sockaddr(sock_addresses[i], RCV_PORT_NUMBER);
	       }
	       par2->sock_addresses = (struct sockaddr **)sock_addresses;
	       par2->max = hosts_num;
	       par2->len = len_s;
	       par2->n_inter = &n_inter;
	       par2->par_rec_c = &par_rec_c;
	       pthread_create(&ptid[n_thread], NULL, &thread_secure_intersection_send_result, par2);
	       n_thread++;
	  }
     }
     /*ciclo: in ascolto finche' non riceve le stringhe in chiaro (l'unione)*/
     while(!end)
     {
	  /*parametri per non rimanere in connect all'infinito*/
	  tv.tv_sec = UNION_CONNECT_TIMEOUT;
	  tv.tv_usec = 0;
	  FD_ZERO(&read_fds);
	  FD_SET(rcv_sockfd, &read_fds);
	  sin_size = sizeof(struct sockaddr_in);
	  if(select(rcv_sockfd + 1, &read_fds, NULL, NULL, &tv) == -1)
	  {
	       printf("[%s]: ", func_name);
	       perror("select");
	       exit(errno);
	  }
	  if(FD_ISSET(rcv_sockfd, &read_fds))/*pronto a leggere dati*/
	  {
	       if((other_sockfd = accept(rcv_sockfd, (struct sockaddr *)sock_addresses[pos], &sin_size)) == -1)
	       {
		    printf("[%s]: ", func_name);
		    perror("accept");
		    exit(errno);
	       }
	       else
	       {
		    SSL* ssl;/*il contesto SSL*/
		    char c_type[MSG_LENGTH];/*che tipo di messaggio:char*/
		    int type;  /*che tipo di messaggio:int*/
		    int byte_rcv;/*quanti bye ho letto*/
		    int n;/*per la accept*/
		    other_sockfds[other_sockfds_num] = other_sockfd;
		    other_sockfds_num++;
		    /*quando chiudero' il socket andra in shutdown senza passare per il  TIME_WAIT*/
		    setsockopt(other_sockfd, SOCK_STREAM, SO_LINGER, (void *)&li, sizeof(struct linger));
		    ssl = SSL_new(ctx_server);
		    SSL_set_fd(ssl, other_sockfd);
		    if((n = SSL_accept(ssl)) < 0)
		    {
			 ssl_get_error(ssl, n, func_name, "SSL_accept");
			 exit(1);
		    }
		    /*leggo il tipo di messagio*/
		    byte_rcv = SSL_read(ssl, c_type, MSG_LENGTH);
		    if(byte_rcv <= 0)
		    {
			 ssl_get_error(ssl, byte_rcv, func_name, "SSL_read");
			 exit(1);
		    }
		    c_type[byte_rcv] = '\0';
		    type = atoi(c_type);
		    /*printf("ricevuta richiesta n[%d]\n", type);*/
		    switch(type)/*switcho il tipo di messaggio*/
		    {
		    case ENCRYPT: /*devo crittare i dati e spedirli al successivo*/
		    {
			 /*parametri da passare ai thread*/
			 struct params_union_intersection_c *par = malloc(sizeof(struct params_union_intersection_c));
			 prepare_sockaddr(sock_addresses[next_pos], RCV_PORT_NUMBER);
			 prepare_sockaddr(sock_addresses[0], RCV_PORT_NUMBER);
			 par->other_sockfd = other_sockfd;
			 par->ssl_server = ssl;
			 par->len = len_s;
			 par->ctx_client = ctx_client;
			 par->sock_address = (struct sockaddr *)sock_addresses[next_pos];
			 par->server_sock_address = (struct sockaddr *)sock_addresses[0];
			 par->par_rec_c = NULL;
			 if(!pos)/*il primo client non invia i messaggi crittati da tutti, se li copia*/
			 {
			      par->server_sock_address = NULL;
			      par->par_rec_c = &par_rec_c;
			 }
			 par->comm_public_key = malloc((strlen(comm_public_key) + 1) * sizeof(char));
			 strcpy(par->comm_public_key, comm_public_key);
			 par->max = hosts_num;
			 par->clear = NULL;
			 par->n_clear = 0;
			 pthread_create(&ptid[n_thread], NULL, 
					&thread_secure_union_intersection_c, par);
			 n_thread++;
			 if(!pos && (protocol == SECURE_INTERSECTION))/*primo host e secure intersection*/
			 {
			      n_encrypt++;
			      /*mi hanno mandato tutti i messaggi posso uscire*/
			      if(n_encrypt == (hosts_num - 1) && n_send_crypt == (hosts_num - 1))
			      {
				   end = 1;
			      }
			 }
			 break;
		    }/*fine case encrypt*/
		    case DECRYPT: /*devo decrittare i dati e spedirli al successivo*/
		    {
			 struct params_union_d *par = malloc(sizeof(struct params_union_d));/*parametri da passare ai thread*/
			 int i;/*per scorrere sock_addresses*/
			 prepare_sockaddr(sock_addresses[next_pos], RCV_PORT_NUMBER);
			 prepare_sockaddr(sock_addresses[0], RCV_PORT_NUMBER);
			 par->ssl_server = ssl;
			 par->len = len_s;
			 par->ctx_client = ctx_client;
			 par->sock_address = (struct sockaddr *)sock_addresses[next_pos];
			 if(pos == (hosts_num - 1)) /*l'ultimo deve spedire i messaggi in chiaro a tutti*/
			 {
			      for(i = 0; i < hosts_num-1; i++)
			      {
				   prepare_sockaddr(sock_addresses[i], RCV_PORT_NUMBER);
			      }
			 }
			 par->server_sock_address = (struct sockaddr **)sock_addresses;/*serve solo all'ultimo*/
			 par->comm_private_key = malloc((strlen(comm_private_key) + 1) * sizeof(char));
			 strcpy(par->comm_private_key, comm_private_key);
			 par->max = hosts_num;
			 par->password = password;
			 par->clear = &final_clear;
			 par->n_clear = &final_n_clear;
			 pthread_create(&ptid[n_thread], NULL, &thread_secure_union_d, par);
			 n_thread++;
			 if(pos == (hosts_num - 1))/*l'ultimo deve uscire*/
			 {
			      end = 1;/*esco (ho inviato i messaggi in chiaro a tutti*/
			 }
			 break;
		    }/*fine case decrypt*/
		    case SEND_RSA: /*arrivano i parametri RSA*/
		    {
			 struct params_rsa_r par;/*parametri da passare ai thread*/
			 pthread_t tid;/*tid del thread*/
			 par.ssl_server = ssl;
			 par.p = &p;
			 par.q = &q;
			 pthread_create(&tid, NULL, 
					&thread_secure_union_intersection_receive_rsa, &par);
			 pthread_join(tid, NULL);
			 my_rsa = RSA_generate_key_commutative(MODULUS_SIZE, *(par.p), *(par.q));
			 if((write_rsa_keys(comm_public_key, comm_private_key, my_rsa, password)) < 0)
			 {
			      printf("Error while writing commutative keys\n");
			      exit(errno);
			 }
			 free(*par.p);
			 free(*par.q);
			 free(my_rsa);
			 len_s = calculate_size(comm_public_key, len_c);
			 if(first)/*la prima volta: almeno i client (eccetto il primo) sanno di inviare i loro dati*/
			 {
			      first = 0;
			      send_first = 1;
			 }
			 break;
		    }/*fine case send_rsa*/
		    case SEND_CRYPT: /*solo per il primo host */
		    {
			 par_rec_c.ssl_server = ssl;
			 pthread_create(&ptid[n_thread], NULL, 
					thread_secure_union_intersection_receive_critt, &par_rec_c);
			 n_thread++;
			 if(!pos && (protocol == SECURE_INTERSECTION))/*primo host e secure intersection*/
			 {
			      n_send_crypt++;
			      if(n_encrypt == (hosts_num - 1) && n_send_crypt == (hosts_num - 1))
			      {
				   /*mi hanno mandato tutti i messaggi posso uscire*/
				   end = 1;
			      }
			 }
			 break;
		    }/*fine send_crypt*/
		    case SEND_CLEAR: /*ricevo l'unione delle stringhe in chiaro*/
		    {
			 struct params_clear_r par;/*parametri da passare ai thread*/
			 pthread_t tid;/*tid del thread*/
			 par.ssl_server = ssl;
			 par.len = len_s;
			 par.clear = &final_clear;
			 par.n_clear = &final_n_clear;
			 pthread_create(&tid, NULL, &thread_secure_union_receive_clear, &par);
			 pthread_join(tid, NULL);
			 end = 1;
			 break;
		    }/*fine send_clear*/
		    case SEND_INTERSECTION: /*ricevo il totale dell'intersezione*/
		    {
			 struct params_intersection_r par;/*parametri da passare ai thread*/
			 pthread_t tid;/*tid del thread*/
			 par.ssl_server = ssl;
			 par.n_inter = &n_inter;
			 pthread_create(&tid, NULL, &thread_secure_intersection_receive_result, &par);
			 pthread_join(tid, NULL);
			 end = 1;
			 break;
		    }/*fine send_intersection*/
		    
		    }/*fine switch*/
		    if(pos && send_first)/*la prima volta critta i propri messaggi in chiaro*/
		    {
			 /*parametri da passare ai thread*/
			 struct params_union_intersection_c *par = malloc(sizeof(struct params_union_intersection_c));
			 send_first = 0;
			 prepare_sockaddr(sock_addresses[next_pos], RCV_PORT_NUMBER);
			 par->other_sockfd = 0;/*non bisogna ricevere*/;
			 par->ssl_server = NULL;
			 par->len = len_s;
			 par->ctx_client = ctx_client;
			 par->sock_address = (struct sockaddr *)sock_addresses[next_pos];
			 par->comm_public_key = malloc((strlen(comm_public_key) + 1) * sizeof(char));
			 strcpy(par->comm_public_key, comm_public_key);
			 par->clear = malloc(n_clear * sizeof(char *));/*alloco i puntatori*/
			 par->max = len_s;
			 par->n_clear = n_clear;
			 for(i = 0; i < n_clear; i++)
			 {
			      par->clear[i] = malloc((len_s+1) * sizeof(char));
			      strcpy(par->clear[i], clear[i]);
			 }
			 pthread_create(&ptid[n_thread], NULL, 
					&thread_secure_union_intersection_c, par);
			 n_thread++;
		    }  
	       }/*else*/
	  }/*if fdisset*/
	  else
	  {
	       printf("Timeout: waiting data from the main thread\n");
	       exit(1);
	  }
     }
     /*aspetto i thread*/
     for(j = 0; j < n_thread; j++)
     {
	  pthread_join(ptid[j], NULL);
     }
     /*chiudo i socket aperti*/
     for(j = 0; j < other_sockfds_num; j++)
     {
	  shutdown(other_sockfds[j], SHUT_RDWR);
	  close(other_sockfds[j]);
     }
     shutdown(rcv_sockfd, SHUT_RDWR);
     close(rcv_sockfd);
     /*il client 0 pulisce le stringhe crittate da tutti*/   
     if(!pos)
     {
	  for(j = 0; j < hosts_num; j++)
	  {
	       int k = 0;
	       while(par_rec_c.critt[j][k] != NULL)
	       {
		    free(par_rec_c.critt[j][k]);
		    k++;
	       }
	       free(par_rec_c.critt[j]);
	  }
	  free(par_rec_c.critt);
     }
     /*pulizia*/
     for(j = 0; j < hosts_num; j++)
     {
	  free(sock_addresses[j]);
     }
     free(sock_addresses);
     SSL_CTX_free(ctx_client);
     SSL_CTX_free(ctx_server);
     ssl_end();
     if(protocol == SECURE_UNION)/*secure union*/
     {
	  *total = final_n_clear;
	  return final_clear;
     }   
     *total = n_inter;

     return NULL;
}


/*esegue il protocollo della secure union tra "hosts_num" host tra gli indirizzi ip 
  ordinati presenti in "ip_addresses", utilizzando l' array "sets" (terminato dal null)
  di elementi di dimensione tutti multipli della stessa chiave (terminati da '\0').
  Ritorna l'unione degli insiemi, scrivendo in "total: la dimensione*/
unsigned char **secure_union(char **ip_addresses, int hosts_num, 
			     char **sets, unsigned long *total)
{
     return secure_union_intersection(ip_addresses, hosts_num, sets, total, SECURE_UNION);
}


/*esegue il protocollo della secure intersection tra "hosts_num" host tra gli indirizzi ip 
  ordinati presenti in "ip_addresses", utilizzando l' array "sets" (terminato dal null)
  di elementi di dimensione tutti multipli della stessa chiave.
  Ritorna la cardinalita' dell'intersezione*/
unsigned long secure_intersection(char **ip_addresses, int hosts_num, char **sets)
{
     unsigned long total;
     secure_union_intersection(ip_addresses, hosts_num, sets, &total, SECURE_INTERSECTION);
     return total;
}


/******************************************************************************/
/*                                                                            */
/*                              SECURE SCALAR PRODUCT                         */
/*                                                                            */
/******************************************************************************/


/*esegue l'elevamento a potenza di "a" elevato alla "b"*/
unsigned long pow(int a, int b)
{
     int i;/*per fare la moltiplicazione*/
     unsigned long c = 1;/*il valore ritornato*/
     for(i = 0; i < b; i++)
     {
	  c *=a;
     }

     return c;
}


/*genera un vettore di "n" unsigned long elementi generati in maniera random*/
unsigned long *generate_random_vector(int n)
{
     unsigned long *vector;/*il vettore ritornato*/
     int i;/*per scorrere il vettore*/

     vector = malloc(n * sizeof(unsigned long));
     for(i = 0; i < n; i++)
     {
	  vector[i] = generate_random_int();
     }

     return vector;
}


/*genera una matrice di rango "n" per "m" unsigned long generati in maniera random*/
unsigned long **generate_random_matrix(int n, int m)
{
     unsigned long **matrix;/*la matrice ritornata*/
     int i;/*per scorrere la matrice*/

     matrix = malloc(n * sizeof(unsigned long *));
     for(i = 0; i < n; i++)
     {
	  matrix[i] = generate_random_vector(m);
     }

     return matrix;
}


/*genera un vettore di "n" unsigned long elementi inizializzati a zero*/
unsigned long *generate_empty_vector(int n)
{
     unsigned long *vector;/*il vettore ritornato*/
     int i;/*per scorrere il vettore*/

     vector = malloc(n * sizeof(unsigned long));
     for(i = 0; i < n; i++)
     {
	  vector[i] = 0;
     }

     return vector;
}


/*genera una matrice di rango "n" per "m" unsigned long inizializzati a zero*/
unsigned long **generate_empty_matrix(int n, int m)
{
     unsigned long **matrix;/*la matrice ritornata*/
     int i;/*per scorrere la matrice*/

     matrix = malloc(n * sizeof(unsigned long *));
     for(i = 0; i < n; i++)
     {
	  matrix[i] = generate_empty_vector(m);
     }

     return matrix;
}


/*calcola e ritorna la trasposta della matrice "ma" di "n" per "m" unsigned long"*/
unsigned long **traspose_matrix(unsigned long **ma, int n, int m)
{
     unsigned long **matrix;/*la matrice ritornata*/
     int i, j;/*per scorrere la matrice*/

     matrix = malloc(m * sizeof(unsigned long *));
     for(i = 0; i < m; i++)
     {
	  matrix[i] = malloc(n * sizeof(unsigned long));
     }
     for(i = 0; i < m; i++)
     {
	  for(j = 0; j < n; j++)
	  {
	       matrix[i][j] = ma[j][i];
	  }
     }

     return matrix;
}


/*esegue il prodotto tra la matrice "ma" e il vettore "v" di dimensioni "m" per "n" ed "n" unsigned long
  e ritorna il risultato sotto forma di vettore*/
unsigned long *mult_matrix_vector(unsigned long **ma, unsigned long *v, int n, int m)
{
     unsigned long *vector;/*il vettore ritornato*/
     int i, j;/*per scorrere gli array*/

     vector = malloc(n * sizeof(unsigned long));
     for(i = 0; i < n; i++)
     {
	  vector[i] = 0;
	  for(j = 0; j < m; j++)
	  {
	       vector[i] += ma[i][j] * v[j];
	  }
     }

     return vector;
}


/*esegue il prodotto tra i vettori "v1" e il vettore "v2" di dimensioni "n" unsigned long
  e ritorna il risultato sotto forma di scalare*/
unsigned long mult_vector_vector(unsigned long *v1, unsigned long *v2, int n)
{
     unsigned long scalar = 0;/*il valore ritornato*/
     int i;/*per scorrere gli array*/

     for(i = 0; i < n; i++)
     {
	  scalar += (v1[i] * v2[i]);
     }

     return scalar;
}


/*esegue la somma tra i vettori "v1" e il vettore "v2" di dimensioni "n" unsigned long
  e ritorna il risultato sotto forma di scalare*/
unsigned long *sum_vector_vector(unsigned long *v1, unsigned long *v2, int n)
{
     unsigned long *vector;/*il vettore ritornato*/
     int i;/*per scorrere il vettore*/

     vector = malloc(n * sizeof(unsigned long));
     for(i = 0; i < n; i++)
     {
	  vector[i] = v1[i] + v2[i];
     }

     return vector;
}


/*funzione per inviare la matrice "matrix" di rango "m" per "n" al sockaddr "sock_address"
  utilizzando "ctx_client"*/
int secure_scalar_send_matrix(SSL_CTX *ctx_client, struct sockaddr *sock_address, 
			      unsigned long **matrix, int m, int n)
{
     SSL *ssl;
     int snd_sockfd;
     char buf[BUF_LENGTH];/*il buffer per mandare i messaggi */
     int len;/*lunghezza del buffer*/
     int i, j;/*per scorrere gli array*/
     int c;/*per la connect*/
     char func_name[] = "secure_scalar_send_matrix";/*il nome della funzione*/

     ssl = SSL_new(ctx_client);
     if((snd_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("socket");
	  exit(errno);
     }
     if((SSL_set_fd(ssl, snd_sockfd)) == 0)
     {
	  print_ssl_error(func_name, "SSL_set_fd");
	  exit(errno);
     }
     if((c = connect_timeout(snd_sockfd, sock_address, SCALAR_CONNECT_TIMEOUT)) < 0)
     {
	  return -1;
     }
     if((c = SSL_connect(ssl)) <= 0)
     {
	  ssl_get_error(ssl, c, func_name, "SSL_connect");
	  exit(errno);	
     }
     /*ciclo di n * m valori*/
     for(i = 0; i < m; i ++)
     {
	  for(j = 0; j < n; j++)
	  {
	       sprintf(buf, "%lu", matrix[i][j]);
	       len = strlen(buf);
	       if((c = SSL_write(ssl, buf, len)) <= 0)
	       {
		    ssl_get_error(ssl, c, func_name, "SSL_write");
		    return -1;
	       }
	  }
     }
     shutdown(snd_sockfd, SHUT_RDWR);
     close(snd_sockfd);
     SSL_free(ssl);
	
     return 0;
}


/*funzione per ricevere una matrice di rango "m" per "n" dal socket "rcv_sockfd" 
  utilizzando la struttura "sock_address" ed il server "ssl", salvando i valori in "matrix"*/
int secure_scalar_receive_matrix(int rcv_sockfd, SSL* ssl, struct sockaddr *sock_address,
				 unsigned long **matrix, int m, int n)
{
     fd_set read_fds;/*per la select*/
     struct timeval tv;/*il timeout per la select*/
     int other_sockfd;/*fd socket dell'altro*/
     char buf[BUF_LENGTH];/*il buffer per ricevere i messaggi */
     int byte_rcv;/*quanti byte ricevuti*/
     socklen_t sin_size;/*dimensione della struct*/
     struct linger li; /*per settare il flag RST dei socket in ascolto(per chiuderli senza stare in TIME_WAIT*/
     int i, j;/*per scorrere gli array**/
     int c;/* per la accept*/
     char func_name[] = "secure_scalar_receive_matrix";/*il nome della funzione*/
     
     li.l_onoff = 1;
     li.l_linger = 0;
     tv.tv_sec = SCALAR_CONNECT_TIMEOUT;
     tv.tv_usec = 0;
     FD_ZERO(&read_fds);
     FD_SET(rcv_sockfd, &read_fds);
     sin_size = sizeof(struct sockaddr_in);
     if(select(rcv_sockfd + 1, &read_fds, NULL, NULL, &tv) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("select");
	  exit(errno);
     }
     if(FD_ISSET(rcv_sockfd, &read_fds))/*pronto a leggere dati*/
     {
	  if((other_sockfd = accept(rcv_sockfd, sock_address, &sin_size)) == -1)
	  {
	       printf("[%s]: ", func_name);
	       perror("accept");
	       exit(errno);
	  }
	  else
	  {
	       /*quando chiudero' il socket andra in shutdown senza passare per il  TIME_WAIT*/
	       setsockopt(other_sockfd, SOCK_STREAM, SO_LINGER, (void *)&li, sizeof(struct linger));
	       SSL_set_fd(ssl, other_sockfd);
	       if((c = SSL_accept(ssl)) < 0)
	       {
		    ssl_get_error(ssl, c, func_name, "SSL_accept");
		    return -1;
	       }
	       for(i = 0; i < m; i++)
	       {
		    for(j = 0; j < n; j++)
		    {
			 byte_rcv = SSL_read(ssl, buf, BUF_LENGTH);
			 if(byte_rcv <= 0)
			 {
			      ssl_get_error(ssl, byte_rcv, func_name, "SSL_read");
			      return -1;
			 }
			 buf[byte_rcv] = '\0';
			 sscanf(buf, "%lu", &matrix[i][j]); /*convert value to a long*/
		    }
	       }
	       shutdown(other_sockfd, SHUT_RDWR);
	       close(other_sockfd);
	  }
     }
     else
     {
	  return -1;
     }
     return 0;
}


/*funzione per inviare il vettore "vector" di lunghezza "n" al sockaddr "sock_address"
  utilizzando "ctx_client"*/
int secure_scalar_send_vector(SSL_CTX *ctx_client, struct sockaddr *sock_address, 
			      unsigned long *vector, int n)
{
     SSL *ssl;
     int snd_sockfd;
     char buf[BUF_LENGTH];/*il buffer per mandare i messaggi */
     int len;/*lunghezza del buffer*/
     int i;/*per scorrere l'array*/
     int c;/*per la connect*/
     char func_name[] = "secure_scalar_send_vector";/*il nome della funzione*/

     ssl = SSL_new(ctx_client);
     if((snd_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("socket");
	  exit(errno);
     }
     if((SSL_set_fd(ssl, snd_sockfd)) == 0)
     {
	  print_ssl_error(func_name, "SSL_set_fd");
	  exit(errno);
     }
     if((c = connect_timeout(snd_sockfd, sock_address, SCALAR_CONNECT_TIMEOUT)) < 0)
     {
	  return -1;
     }
     if((c = SSL_connect(ssl)) <= 0)
     {
	  ssl_get_error(ssl, c, func_name, "SSL_connect");
	  exit(errno);	
     }
     /*ciclo di n valori*/
     for(i = 0; i < n; i ++)
     {
	  sprintf(buf, "%lu", vector[i]);
	  len = strlen(buf);
	  if((c = SSL_write(ssl, buf, len)) <= 0)
	  {
	       ssl_get_error(ssl, c, func_name, "SSL_write");
	       return -1;
	  }
     }
     shutdown(snd_sockfd, SHUT_RDWR);
     close(snd_sockfd);
     SSL_free(ssl);

     return 0;
}


/*funzione per ricevere un vettore di lunghezza "n" dal socket "rcv_sockfd" 
  utilizzando la struttura "sock_address" ed il server "ssl", salvando i valori in "vector"*/
int secure_scalar_receive_vector(int rcv_sockfd, SSL* ssl, struct sockaddr *sock_address,
				 unsigned long *vector, int n)
{
     fd_set read_fds;/*per la select*/
     struct timeval tv;/*il timeout per la select*/
     int other_sockfd;/*fd socket dell'altro*/
     char buf[BUF_LENGTH];/*il buffer per ricevere i messaggi */
     int byte_rcv;/*quanti byte ricevuti*/
     socklen_t sin_size;/*dimensione della struct*/
     struct linger li; /*per settare il flag RST dei socket in ascolto(per chiuderli senza stare in TIME_WAIT*/
     int i;/*per scorrere l'array*/
     int c;/*per la accept*/
     char func_name[] = "secure_scalar_receive_vector";/*il nome della funzione*/

     li.l_onoff = 1;
     li.l_linger = 0;
     tv.tv_sec = SCALAR_CONNECT_TIMEOUT;
     tv.tv_usec = 0;
     FD_ZERO(&read_fds);
     FD_SET(rcv_sockfd, &read_fds);
     sin_size = sizeof(struct sockaddr_in);
     if(select(rcv_sockfd + 1, &read_fds, NULL, NULL, &tv) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("select");
	  exit(errno);
     }
     if(FD_ISSET(rcv_sockfd, &read_fds))/*pronto a leggere dati*/
     {
	  if((other_sockfd = accept(rcv_sockfd, sock_address, &sin_size)) == -1)
	  {
	       printf("[%s]: ", func_name);
	       perror("accept");
	       exit(errno);
	  }
	  else
	  {

	       /*quando chiudero' il socket andra in shutdown senza passare per il  TIME_WAIT*/
	       setsockopt(other_sockfd, SOCK_STREAM, SO_LINGER, (void *)&li, sizeof(struct linger));
	       SSL_set_fd(ssl, other_sockfd);
	       if((c = SSL_accept(ssl)) < 0)
	       {
		    ssl_get_error(ssl, c, func_name, "SSL_accept");
		    return -1;
	       }
	       for(i = 0; i < n; i++)
	       {
		    byte_rcv = SSL_read(ssl, buf, BUF_LENGTH);
		    if(byte_rcv <= 0)
		    {
			 ssl_get_error(ssl, byte_rcv, func_name, "SSL_read");
			 return -1;
		    }
		    buf[byte_rcv] = '\0';
		    sscanf(buf, "%lu", &vector[i]); /*convert value to a long*/
	       }
	       shutdown(other_sockfd, SHUT_RDWR);
	       close(other_sockfd);
	  }
     }
     else
     {
	  return -1;
     }

     return 0;
}


/*funzione per inviare lo scalare "scalar" al sockaddr "sock_address"
  utilizzando "ctx_client"*/
int secure_scalar_send_scalar(SSL_CTX *ctx_client, struct sockaddr *sock_address, 
			      unsigned long scalar)
{
     SSL *ssl;/*serve per inviare i dati*/
     int snd_sockfd;/*il sending socket*/
     char buf[BUF_LENGTH];/*il buffer per mandare i messaggi */
     int len;/*lunghezza del buffer*/
     int c;/*per la connect*/
     char func_name[] = "secure_scalar_send_scalar";/*il nome della funzione*/

     ssl = SSL_new(ctx_client);
     if((snd_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("socket");
	  exit(errno);
     }
     if((SSL_set_fd(ssl, snd_sockfd)) == 0)
     {
	  print_ssl_error(func_name, "SSL_set_fd");
	  exit(errno);
     }
     if((c = connect_timeout(snd_sockfd, sock_address, SCALAR_CONNECT_TIMEOUT)) < 0)
     {
	  return -1;
     }
     if((c = SSL_connect(ssl)) <= 0)
     {
	  ssl_get_error(ssl, c, func_name, "SSL_connect");
	  exit(errno);	
     }
     sprintf(buf, "%lu", scalar);
     len = strlen(buf);
     if((c = SSL_write(ssl, buf, len)) <= 0)
     {
	  ssl_get_error(ssl, c, func_name, "SSL_write");
	  return -1;
     }
     shutdown(snd_sockfd, SHUT_RDWR);
     close(snd_sockfd);
     SSL_free(ssl);
	
     return 0;
}


/*funzione per ricevere uno scalare dal socket "rcv_sockfd" 
  utilizzando la struttura "sock_address" ed il server "ssl", salvando il valore in "scalar"*/
int secure_scalar_receive_scalar(int rcv_sockfd, SSL* ssl, struct sockaddr *sock_address,
				 unsigned long *scalar)
{
     fd_set read_fds;/*per la select*/
     struct timeval tv;/*il timeout per la select*/
     int other_sockfd;/*fd socket dell'altro*/
     char buf[BUF_LENGTH];/*il buffer per ricevere i messaggi */
     int byte_rcv;/*quanti byte ricevuti*/
     socklen_t sin_size;/*dimensione della struct*/
     struct linger li; /*per settare il flag RST dei socket in ascolto(per chiuderli senza stare in TIME_WAIT*/
     int c;/*per la accept*/
     char func_name[] = "secure_scalar_receive_scalar";/*il nome della funzione*/

     li.l_onoff = 1;
     li.l_linger = 0;
     tv.tv_sec = SCALAR_CONNECT_TIMEOUT;
     tv.tv_usec = 0;
     FD_ZERO(&read_fds);
     FD_SET(rcv_sockfd, &read_fds);
     sin_size = sizeof(struct sockaddr_in);
     if(select(rcv_sockfd + 1, &read_fds, NULL, NULL, &tv) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("select");
	  exit(errno);
     }
     if(FD_ISSET(rcv_sockfd, &read_fds))/*pronto a leggere dati*/
     {
	  if((other_sockfd = accept(rcv_sockfd, sock_address, &sin_size)) == -1)
	  {
	       printf("[%s]: ", func_name);
	       perror("accept");
	       exit(errno);
	  }
	  else
	  {
	       /*quando chiudero' il socket andra in shutdown senza passare per il  TIME_WAIT*/
	       setsockopt(other_sockfd, SOCK_STREAM, SO_LINGER, (void *)&li, sizeof(struct linger));
	       SSL_set_fd(ssl, other_sockfd);
	       if((c = SSL_accept(ssl)) < 0)
	       {
		    ssl_get_error(ssl, c, func_name, "SSL_accept");
		    return -1;
	       }
	       byte_rcv = SSL_read(ssl, buf, BUF_LENGTH);
	       if(byte_rcv <= 0)
	       {
		    ssl_get_error(ssl, byte_rcv, func_name, "SSL_read");
		    return -1;
	       }
	       buf[byte_rcv] = '\0';
	       sscanf(buf, "%lu", scalar); /*convert value to a long*/
	  }
	  shutdown(other_sockfd, SHUT_RDWR);
	  close(other_sockfd);
     }
     else
     {
	  return -1;
     }

     return 0;
}


/*esegue il protocollo del secure scalar product tra gli indirizzi ip ordinati 
  presenti in "ip_addresses"; "set" rappresenta una colonna (A e B si calcolano 
  prima set come prodotto delle loro colonne facenti parte dell'insieme);
  "hosts_num" (il numero di indirizzi presenti in ip_addresses) deve essere uguale a 2*/
unsigned long secure_scalar_product(char **ip_addresses, int hosts_num, char *set)
{

     psockaddr_in *sock_addresses;/*puntatori alle strutture sock_add*/
     int rcv_sockfd;/*fd socket d'ascolto*/
     int pos = 0, next_pos;/*la posizione dell'ip e del successivo*/	
     char my_name[NAME_LENGTH];/*contiene il nome del client*/
     char my_ip[MAX_IP_LENGTH];/*indirizzo ip mio*/	
     SSL *ssl_server;/*gli oggetti ssl in invio ed ascolto*/
     SSL_CTX *ctx_client, *ctx_server;/*i context ssl lato client e server*/
     char* CAfile = CAFILE;/*nome del file con i certificati pubblici in formato testo*/
     char cert[MAX_FILENAME_LENGTH] = CERTIFICATE_DIR;/*nome del file con il certificato del client*/
     char secret_key[MAX_FILENAME_LENGTH] = CERTIFICATE_DIR;/*nome del file con la chiave private del client*/
     struct linger li; /*per settare il flag RST dei socket in ascolto (per chiuderli senza stare in TIME_WAIT)*/
     /*variabili specifiche del scalar product: stessi nomi della pubblicazione*/
     int n; /*dimensione della colonna (lunghezza)*/
     int m; /*n/2*/
     int r; /*r minore di n*/
     int prop = N_R_PROPORTION; /*calcolo r = n/prop */
     unsigned long *vector_r; /*serve ad A*/
     unsigned long *vector_r1;/*serve a B*/
     unsigned long *vector_r2 = malloc(r * sizeof(unsigned long)); /*serve ad A e b*/
     unsigned long *vector_x; /*set per A*/
     unsigned long *vector_y; /*set per B*/
     unsigned long **half_up_matrix;/*serve ad A e B*/
     unsigned long **half_down_matrix;/*serve ad A e B*/
     unsigned long **c_matrix;/*la matrice di comune accordo*/
     unsigned long **c_t_matrix;/*la matrice trasposta*/
     unsigned long *vector_x1;/*serve ad A*/
     unsigned long *vector_x2;/*serve ad A e B*/
     unsigned long *vector_y1;/*serve a B e A*/
     unsigned long scalar_s1;/*serve a B e A*/
     unsigned long scalar_s2;/*serve a B e A*/
     unsigned long scalar_s3;/*serve a B*/
     unsigned long final_result;/*serve ad A e B*/
     int i, j, k;/*per scorrere gli array*/
     char func_name[] = "secure_scalar_product";/*il nome della funzione*/

     /*calcolo la lunghezza della colonna*/
     n = strlen(set);
     r = n/prop;
     m = n/2;     
     li.l_onoff = 1;
     li.l_linger = 0;
     ssl_init();/*inizializzo ssl*/
     gethostname(my_name, NAME_LENGTH);	
     sock_addresses = create_sockaddresses(ip_addresses, hosts_num, my_name, my_ip, &pos);
     /*solo tra due host*/
     if(hosts_num !=  2)
     {
	  printf("Scalar product can only be executed between two hosts.\n");
	  exit(1);
     }
     /*adesso comincia il protocollo di comunicazione*/
     next_pos = (pos + 1) % hosts_num;
     /*preparo le stringhe per i file*/
     strcat(cert, my_ip);
     strcat(cert, CERTIFICATE_EXT);
     strcat(secret_key, my_ip);
     strcat(secret_key, PRIVATE_EXT);
     /*creo i contesti client e server*/
     if((ctx_client = ssl_setup_ctx_client(CAfile, cert, secret_key)) == NULL)
     {
	  exit(errno);
     }
     if((ctx_server = ssl_setup_ctx_server(CAfile, cert, secret_key)) == NULL)
     {
	  exit(errno);
     }
     /*preparo il receiving socket*/
     if((rcv_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("socket");
	  exit(errno);
     }
     prepare_sockaddr(sock_addresses[pos], SCALAR_PORT_NUMBER);
     if(bind(rcv_sockfd, (struct sockaddr *)sock_addresses[pos], sizeof(struct sockaddr)) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("bind");
	  exit(errno);
     }   
     if(listen(rcv_sockfd, 1) == -1)
     {
	  printf("[%s]: ", func_name);
	  perror("listen");
	  exit(errno);
     }
     /*non sta in TIME_WAIT dopo la close*/
     setsockopt(rcv_sockfd, SOCK_STREAM, SO_LINGER, (void *)&li, sizeof(struct linger));
     if(!pos)/*sono A*/
     {
	  /*il vettore di A*/
	  vector_x = generate_empty_vector(n);/*vettore x dimensione n*/
	  for(i = 0; i < n; i ++)
	  {
	       /*strcmp ritorna 0 (false) se le stringhe sono uguali*/
	       vector_x[i] = strncmp(&set[i], "0", sizeof(char)) ? 1 : 0;
	  }

	  half_up_matrix = generate_random_matrix(m, n);/*parte superiore di C, dim m * n */
	  /*preparo il sock_addr dell'altro*/
	  prepare_sockaddr(sock_addresses[next_pos], SCALAR_PORT_NUMBER);
	  /*invio la parte superiore di c matrix*/
	  if(secure_scalar_send_matrix(ctx_client, (struct sockaddr *)sock_addresses[next_pos], 
				       half_up_matrix, m, n) != 0)
	  {
	       printf("Impossible to connect to %s\n", inet_ntoa(sock_addresses[next_pos]->sin_addr));
	       exit(errno);
	  }
	  half_down_matrix = generate_empty_matrix((n-m), n);/*parte inferiore di C, dim (n-m) * n */
	  /*preparo il server SSL*/
	  if((ssl_server = SSL_new(ctx_server)) == NULL)
	  {
	       print_ssl_error(func_name, "SSL_new");
	       exit(1);
	  }
	  /*ricevo la parte inferiore di c matrix*/
	  if(secure_scalar_receive_matrix(rcv_sockfd, ssl_server,  (struct sockaddr *)sock_addresses[next_pos],
					  half_down_matrix, (n-m), n) != 0)
	  {
	       printf("Timeout: waiting for data from %s\n", 
		      inet_ntoa(sock_addresses[(next_pos)]->sin_addr));
	       exit(1);
	  }
	  SSL_free(ssl_server);
	  /*adesso posso costruire c matrix*/
	  c_matrix = generate_empty_matrix(n, n);/*matrice C: n*n */
	  for(i = 0; i < m; i++)
	  {
	       for(j = 0; j < n; j++)
	       {
		    c_matrix[i][j] = half_up_matrix[i][j];
	       }
	  }
	  for(i = 0; i < (n-m); i++)
	  {
	       for(j = 0; j < n; j++)
	       {
		    c_matrix[i+m][j] = half_down_matrix[i][j];
	       }
	  }
	  

	  /*eseguo i calcoli*/
	  vector_r = generate_random_vector(n); /*vettore r: random di dimensione n*/
	  vector_x1 = mult_matrix_vector(c_matrix, vector_r, n, n); /*vettore x1 = C x R, dim. n */
	  vector_x2 = sum_vector_vector(vector_x, vector_x1, n); /* vettore x2 = x + x1 */
	  prepare_sockaddr(sock_addresses[next_pos], SCALAR_PORT_NUMBER);
	  /*invio il vettore x2 */
	  if(secure_scalar_send_vector(ctx_client, (struct sockaddr *)sock_addresses[next_pos], 
				       vector_x2, n) != 0)
	  {
	       printf("Impossible to connect to %s\n", inet_ntoa(sock_addresses[next_pos]->sin_addr));
	       exit(errno);
	  }
	  if((ssl_server = SSL_new(ctx_server)) == NULL)
	  {
	       print_ssl_error(func_name, "SSL_new");
	       exit(1);
	  }
	  /*ricevo s1*/
	  if(secure_scalar_receive_scalar(rcv_sockfd, ssl_server,  (struct sockaddr *)sock_addresses[next_pos],
					  &scalar_s1) != 0)
	  {
	       printf("Timeout: waiting for data from %s\n", 
		      inet_ntoa(sock_addresses[(next_pos)]->sin_addr));
	       exit(1);
	  }
	  SSL_free(ssl_server);
	  vector_y1 = generate_empty_vector(n);/*vettore y1, dimensione n*/
	  if((ssl_server = SSL_new(ctx_server)) == NULL)
	  {
	       print_ssl_error(func_name, "SSL_new");
	       exit(1);
	  }
	  /*ricevo y1*/
	  if(secure_scalar_receive_vector(rcv_sockfd, ssl_server,  (struct sockaddr *)sock_addresses[next_pos],
					  vector_y1, n) != 0)
	  {
	       printf("Timeout: waiting for data from %s\n", 
		      inet_ntoa(sock_addresses[(next_pos)]->sin_addr));
	       exit(1);
	  }
	  SSL_free(ssl_server);
	  /*faccio i calcoli: r2[0] = somma i primi r/n valori di r , r2[1] = etc..*/
	  scalar_s2 = mult_vector_vector(vector_r, vector_y1, n); /*scalare s2 = R * Y1 */
	  vector_r2 = generate_empty_vector(r);/*vettore r2, dimensione r */
	  j = 0;
	  k = 1;
	  for(i = 0; i < r; i++)
	  {
	       vector_r2[i] = 0;
	  }
	  
	  for(i = 0; i < n; i++)
	  {
	       vector_r2[j] += vector_r[i];
	       if((i + 1) >= k * (n/r))
	       {
		    if(k < r )
		    {
			 k++;
			 j++;
		    }
	       }
	  }
	  prepare_sockaddr(sock_addresses[next_pos], SCALAR_PORT_NUMBER);
	  /*invio r2*/
	  if(secure_scalar_send_vector(ctx_client, (struct sockaddr *)sock_addresses[next_pos], 
				       vector_r2, r) != 0)
	  {
	       printf("Impossible to connect to %s\n", inet_ntoa(sock_addresses[next_pos]->sin_addr));
	       exit(errno);
	  }
	  prepare_sockaddr(sock_addresses[next_pos], SCALAR_PORT_NUMBER);
	  /*invio s2*/
	  if(secure_scalar_send_scalar(ctx_client, (struct sockaddr *)sock_addresses[next_pos], 
				       scalar_s2) != 0)
	  {
	       printf("Impossible to connect to %s\n", inet_ntoa(sock_addresses[next_pos]->sin_addr));
	       exit(errno);
	  }
	  if((ssl_server = SSL_new(ctx_server)) == NULL)
	  {
	       print_ssl_error(func_name, "SSL_new");
	       exit(1);
	  }
	  /*ricevo il risultato finale*/
	  if(secure_scalar_receive_scalar(rcv_sockfd, ssl_server,  (struct sockaddr *)sock_addresses[next_pos],
					  &final_result) != 0)
	  {
	       printf("Timeout: waiting for data from %s\n", 
		      inet_ntoa(sock_addresses[(next_pos)]->sin_addr));
	       exit(1);
	  }
	  SSL_free(ssl_server);
	  /*pulizia vettori solo di A*/
	  free(vector_x);
	  free(vector_r);
	  free(vector_x1);
     }
     else/*sono B*/
     {
	  vector_y = generate_empty_vector(n);/*vettore y, dimensione n*/
	  for(i = 0; i < n; i ++)
	  {
	       vector_y[i] = strncmp(&set[i], "0", sizeof(char)) ? 1 : 0;
	  }
	  half_up_matrix = generate_empty_matrix(m, n);/*parte superiore di C, dim m * n */
	  /*preparo il server SSL*/
	  if((ssl_server = SSL_new(ctx_server)) == NULL)
	  {
	       print_ssl_error(func_name, "SSL_new");
	       exit(1);
	  }
	  /*ricevo la parte superiore della matrice*/
	  if(secure_scalar_receive_matrix(rcv_sockfd, ssl_server,  (struct sockaddr *)sock_addresses[next_pos],
					  half_up_matrix, m, n) != 0)
	  {
	       printf("Timeout: waiting for data from %s\n", 
		      inet_ntoa(sock_addresses[(next_pos)]->sin_addr));
	       exit(1);
	  }
	  SSL_free(ssl_server);
	  half_down_matrix = generate_random_matrix((n-m), n);/*parte inferiore di C, dim (n-m) * n */
	  prepare_sockaddr(sock_addresses[next_pos], SCALAR_PORT_NUMBER);
	  /*invio la parte inferiore della matrice*/
	  if(secure_scalar_send_matrix(ctx_client, (struct sockaddr *)sock_addresses[next_pos], 
				       half_down_matrix, (n-m), n) != 0)
	  {
	       printf("Impossible to connect to %s\n", inet_ntoa(sock_addresses[next_pos]->sin_addr));
	       exit(errno);
	  }
	  /*adesso posso costruire c matrix*/
	  c_matrix = generate_empty_matrix(n, n);/*matrice C: n*n */
	  for(i = 0; i < m; i++)
	  {
	       for(j = 0; j < n; j++)
	       {
		    c_matrix[i][j] = half_up_matrix[i][j];
	       }
	  }
	  for(i = 0; i < (n-m); i++)
	  {
	       for(j = 0; j < n; j++)
	       {
		    c_matrix[i+m][j] = half_down_matrix[i][j];
	       }
	  }
     

	  vector_x2 = generate_empty_vector(n);
	  if((ssl_server = SSL_new(ctx_server)) == NULL)
	  {
	       print_ssl_error(func_name, "SSL_new");
	       exit(1);
	  }
	  /*ricevo x2*/
	  if(secure_scalar_receive_vector(rcv_sockfd, ssl_server,  (struct sockaddr *)sock_addresses[next_pos],
					  vector_x2, n) != 0)
	  {
	       printf("Timeout: waiting for data from %s\n", 
		      inet_ntoa(sock_addresses[(next_pos)]->sin_addr));
	       exit(1);
	  }
	  SSL_free(ssl_server);
	  /*faccio i calcoli*/
	  c_t_matrix = traspose_matrix(c_matrix, n, n);
	  scalar_s1 = mult_vector_vector(vector_x2, vector_y, n); /* scalare s1 = x2 * Y */
	  vector_y1 = mult_matrix_vector(c_t_matrix, vector_y, n, n); /* vettore y1 = Ct * Y */
	  vector_r1 = generate_random_vector(r);/* vettore r1 random dimension r*/
	  /*faccio i calcoli: y1[0] = somma i primi r/n valori di r1 , y1[1] = etc..*/
	  j = 0;
	  k = 1;
	  for(i = 0; i < n; i++)
	  {
	       vector_y1[i] += vector_r1[j];
	       if((i + 1) >= k * (n/r)) /*[1.... n/r];  [(n/r) + 1....2*(n/r)];...;[(((r-1)/n)/r) + 1...n] */
	       {
		    if(k < r)/*l'ultima volta arrivo fino ad n*/
		    {
			 k++;
			 j++;
		    }
	       }
	  }
	  prepare_sockaddr(sock_addresses[next_pos], SCALAR_PORT_NUMBER);
	  /*invio s1*/
	  if(secure_scalar_send_scalar(ctx_client, (struct sockaddr *)sock_addresses[next_pos], 
				       scalar_s1) != 0)
	  {
	       printf("Impossible to connect to %s\n", inet_ntoa(sock_addresses[next_pos]->sin_addr));
	       exit(errno);
	  }
	  prepare_sockaddr(sock_addresses[next_pos], SCALAR_PORT_NUMBER);
	  /*invio y1*/
	  if(secure_scalar_send_vector(ctx_client, (struct sockaddr *)sock_addresses[next_pos], 
				       vector_y1, n) != 0)
	  {
	       printf("Impossible to connect to %s\n", inet_ntoa(sock_addresses[next_pos]->sin_addr));
	       exit(errno);
	  }
	  vector_r2 = generate_empty_vector(r);
	  if((ssl_server = SSL_new(ctx_server)) == NULL)
	  {
	       print_ssl_error(func_name, "SSL_new");
	       exit(1);
	  }
	  /*ricevo r2*/
	  if(secure_scalar_receive_vector(rcv_sockfd, ssl_server,  (struct sockaddr *)sock_addresses[next_pos],
					  vector_r2, r) != 0)
	  {
	       printf("Timeout: waiting for data from %s\n", 
		      inet_ntoa(sock_addresses[(next_pos)]->sin_addr));
	       exit(1);
	  }
	  SSL_free(ssl_server);
	  if((ssl_server = SSL_new(ctx_server)) == NULL)
	  {
	       print_ssl_error(func_name, "SSL_new");
	       exit(1);
	  }
	  /*ricevo s2*/
	  if(secure_scalar_receive_scalar(rcv_sockfd, ssl_server,  (struct sockaddr *)sock_addresses[next_pos],
					  &scalar_s2) != 0)
	  {
	       printf("Timeout: waiting for data from %s\n", 
		      inet_ntoa(sock_addresses[(next_pos)]->sin_addr));
	       exit(1);
	  }
	  SSL_free(ssl_server);
	  /*faccio i calcoli*/
	  scalar_s3 = mult_vector_vector(vector_r1, vector_r2, r); /*scalare s3 = R1 * R2 */
	  final_result = scalar_s1 - scalar_s2 + scalar_s3;
	  prepare_sockaddr(sock_addresses[next_pos], SCALAR_PORT_NUMBER);
	  /*invio il risultato finale*/
	  if(secure_scalar_send_scalar(ctx_client, (struct sockaddr *)sock_addresses[next_pos], 
				       final_result) != 0)
	  {
	       printf("Impossible to connect to %s\n", inet_ntoa(sock_addresses[next_pos]->sin_addr));
	       exit(errno);
	  }
	  /* free vector e matrici solo B */
	  for(i = 0; i < n; i++)
	  {
	       free(c_t_matrix[i]);
	  }
	  free(c_t_matrix);
	  free(vector_y);
	  free(vector_r1);
     }
     shutdown(rcv_sockfd, SHUT_RDWR);
     close(rcv_sockfd);
     /*free variabili comuni*/
     for(i = 0; i < m; i++)
     {
	  free(half_up_matrix[i]);
     }
     free(half_up_matrix);
     for(i = 0; i < (n-m); i++)
     {
	  free(half_down_matrix[i]);
     }
     free(half_down_matrix);
     for(i = 0; i < n; i++)
     {
	  free(c_matrix[i]);
     }
     free(c_matrix);
     free(vector_x2);
     free(vector_y1);
     free(vector_r2);     
     SSL_CTX_free(ctx_client);
     SSL_CTX_free(ctx_server);
     ssl_end();

     return final_result;
}
