/******************************************************************************/
/*file ssl.c                      SSL functions                               */
/*                          coded by Daniele Sgandurra                        */
/******************************************************************************/

#include "ssl.h"
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <limits.h>
#include <sys/types.h>
#include <openssl/ssl.h>
#include <openssl/x509.h>
#include <openssl/rand.h>
#include <openssl/err.h>
#include <openssl/evp.h>


/*stampa a video i messaggi di errore, della funzione di OpenSSL "ssl_func" chiamata 
 all'interno della funzione "func_name"*/
void print_ssl_error(char *func_name, char *ssl_func)
{
     int err = ERR_get_error();
     printf("[%s]: %s: %s\n", func_name, ssl_func, ERR_error_string(err, NULL));
     ERR_free_strings();
}


/*controlla qual e' l'errore nella struttura "ssl", dopo una chiamata alla funzione
  "ssl_func" di OpenSSL che ha ritornato il valore "ret", eseguita all'interno della
  funzione "func_name"*/
void ssl_get_error(SSL *ssl, int ret, char* func_name, char *ssl_func)
{
     int err = SSL_get_error(ssl, ret);
     switch(err)
     {
     case SSL_ERROR_NONE:
	  print_ssl_error(func_name, ssl_func);
	  break;
     case SSL_ERROR_SSL:
	  print_ssl_error(func_name, ssl_func);
	  break;
     case SSL_ERROR_WANT_READ:
	  print_ssl_error(func_name, ssl_func);
	  break;
     case SSL_ERROR_WANT_WRITE:
	  print_ssl_error(func_name, ssl_func);
	  break;
     case SSL_ERROR_WANT_X509_LOOKUP:
	  print_ssl_error(func_name, ssl_func);
	  break;
     case SSL_ERROR_SYSCALL:
     {
	  err = ERR_get_error();
	  if(err <= 0)
	  {
	       printf("[%s]: \n", func_name);
	       perror(ssl_func);
	       exit(errno);
	  }
	  else
	  {
	       printf("[%s]: %s: %s\n", func_name, ssl_func, ERR_error_string(err, NULL));
	       ERR_free_strings();
	  }
     }
     break;
     case SSL_ERROR_ZERO_RETURN:
	  print_ssl_error(func_name, ssl_func);
	  break;
     case SSL_ERROR_WANT_CONNECT:
	  print_ssl_error(func_name, ssl_func);
	  break;
     default:
	  print_ssl_error(func_name, ssl_func);
	  break;
     }
}
  

/*inizializza OpenSSL*/
void ssl_init()
{

     SSL_load_error_strings();
     ERR_load_crypto_strings();
     SSL_library_init();
     OpenSSL_add_all_algorithms();
     OpenSSL_add_all_ciphers();
     OpenSSL_add_all_digests();   
}


/*libera la memoria dalle stringhe e dagli algoritmi caricati da OpenSSL*/
void ssl_end()
{
     ERR_free_strings();
     EVP_cleanup();
}


/* crea il contesto per la connessione SSL lato client, utilizzando
   il certificato presente in "certfile" contenente la lista di tutti
   gli host, il certificato del client presente in "certificate", e la 
   chiave private presente in "privkey"*/
SSL_CTX* ssl_setup_ctx_client(char *certfile, char *certificate, char *privkey)
{
     SSL_CTX *ctx; /*il contesto*/
     char func_name[] = "ssl_setup_ctx_client";/*il nome della funzione*/

     /*creo il context*/
     if((ctx = SSL_CTX_new(SSLv3_client_method())) == NULL)
     {
	  print_ssl_error(func_name, "SSL_CTX_new");
	  return NULL;
     }
     
     /*uso il certificato*/
     SSL_CTX_load_verify_locations(ctx, certfile, NULL);
     SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, NULL);
     /*richiedo di utilizzare il certificato*/
     if(!SSL_CTX_use_certificate_file(ctx, certificate, SSL_FILETYPE_PEM))
     {
	  print_ssl_error(func_name, "SSL_CTX_use_certificate_file");
	  return NULL;
     }
     /*uso la chiave privata*/
     if(!SSL_CTX_use_PrivateKey_file(ctx, privkey, SSL_FILETYPE_PEM))
     {	
	  print_ssl_error(func_name, "SSL_CTX_use_PrivateKey_file");		
	  return NULL;
     }
     return ctx;
}


/* crea il contesto per la connessione SSL lato server, utilizzando
   il certificato presente in "certfile" contenente la lista di tutti
   gli host, il certificato del client presente in "certificate", e la 
   chiave private presente in "privkey"*/
SSL_CTX *ssl_setup_ctx_server(char *certfile, char *certificate, char *privkey)
{
     SSL_CTX *ctx; /*il contesto*/
     char func_name[] = "ssl_setup_ctx_server";/*il nome della funzione*/

     if((ctx = SSL_CTX_new(SSLv3_server_method())) == NULL)
     {	
	  print_ssl_error(func_name, "SSL_CTX_new");
	  return NULL;
     }
     SSL_CTX_load_verify_locations(ctx, certfile, NULL);
     SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, NULL);

     if(!SSL_CTX_use_certificate_file(ctx, certificate, SSL_FILETYPE_PEM))
     {
	  print_ssl_error(func_name, "SSL_CTX_use_certificate_file");
	  return NULL;
     }

     if(!SSL_CTX_use_PrivateKey_file(ctx, privkey, SSL_FILETYPE_PEM))
     {
	  print_ssl_error(func_name, "SSL_CTX_use_PrivateKey_file");
	  return NULL;
     }
     return ctx;
}
